#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import itertools
import pathlib
import pickle
import tempfile

import h5py
import hdf5storage
import joblib as jl
import numpy as np
import openpyxl
import pandas as pd
import pytest
import scipy.io
import torch

import cebra.data.load as cebra_load

__test_functions = []
__test_functions_error = []
__test_functions_module_not_found = []


def _skip_hdf5storage(*args, **kwargs):
    pytest.skip(reason=(
        "Likely upstream issue with hdf5storage. "
        "For details, see https://github.com/stes/neural_cl/issues/417"))(
            *args, **kwargs)


def test_imports():
    import cebra

    assert hasattr(cebra, "load_data")


def register(*file_endings, requires=()):
    # for each file format
    def _register(f):
        # f is the filename
        # TODO: test loading for file without extension
        # __test_functions.append(f)
        # with all possible extensions
        __test_functions.extend([(lambda filename, dtype: f(
            filename + "." + file_ending, dtype=dtype), file_ending)
                                 for file_ending in file_endings])
        if len(requires) > 0:
            __test_functions_module_not_found.extend([
                (requires, lambda filename: filename + "." + file_ending, lambda
                 filename, dtype: f(filename + "." + file_ending, dtype=dtype))
                for file_ending in file_endings
            ])
        return f

    return _register


def register_error(*file_endings):
    # for each file format
    def _register(f):
        # f is the filename
        # TODO: test loading for file without extension
        # __test_functions_error.append(f)
        # with all possible extensions
        __test_functions_error.extend([
            lambda filename: f(filename + "." + file_ending)
            for file_ending in file_endings
        ])
        return f

    return _register


##### .NPY #####
@register("npy")
def generate_numpy(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    np.save(filename, A)
    loaded_A = cebra_load.load(filename)
    return A, loaded_A


@register("npy")
def generate_numpy_path(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    np.save(filename, A)
    loaded_A = cebra_load.load(pathlib.Path(filename))
    return A, loaded_A


# def test_load_numpy_pickle():
#     assert False


##### . NPZ #####
@register("npz")
def generate_numpy_confounder(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    np.savez(filename, array=A, other_data="test")
    loaded_A = cebra_load.load(filename)
    return A, loaded_A


@register("npz")
def generate_numpy_path_2(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    np.savez(filename, array=A, other_data="test")
    loaded_A = cebra_load.load(pathlib.Path(filename))
    return A, loaded_A


@register("npz")
def generate_numpy_key(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    np.savez(filename, array=A, other_data="test")
    loaded_A = cebra_load.load(filename, key="array")
    return A, loaded_A


@register("npz")
def generate_numpy_second(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    np.savez(filename, other_data="test", array=A)
    loaded_A = cebra_load.load(filename)
    return A, loaded_A


@register_error("npz")
def generate_numpy_wrong_key(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    np.savez(filename, array=A, other_data="test")
    _ = cebra_load.load(filename, key="wrong_array")


@register_error("npz")
def generate_numpy_invalid_key(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    np.savez(filename, array=A, other_data="test")
    _ = cebra_load.load(filename, key="other_data")


@register_error("npz")
def generate_numpy_no_array(filename, dtype):
    np.savez(filename, array="test_1", other_data="test_2")
    _ = cebra_load.load(filename)


#### .H5 #####
# TODO: test raise ModuleFoundError for h5py


@register("h5", "hdf", "hdf5", "h", requires=("h5py",))
def generate_h5(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    with h5py.File(filename, "w") as hf:
        hf.create_dataset("dataset_1", data=A)
    loaded_A = cebra_load.load(filename)
    return A, loaded_A


@register("h5", "hdf", "hdf5", "h")
def generate_h5_confounder(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    with h5py.File(filename, "w") as hf:
        hf.create_dataset("dataset_1", data=A)
        hf.create_dataset("dataset_2", data="test")
    loaded_A = cebra_load.load(filename)
    return A, loaded_A


@register("h5", "hdf", "hdf5", "h")
def generate_h5_path(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    with h5py.File(filename, "w") as hf:
        hf.create_dataset("dataset_1", data=A)
        hf.create_dataset("dataset_2", data="test")
    loaded_A = cebra_load.load(pathlib.Path(filename))
    return A, loaded_A


@register("h5", "hdf", "hdf5", "h")
def generate_h5_key(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    with h5py.File(filename, "w") as hf:
        hf.create_dataset("dataset_1", data=A)
        hf.create_dataset("dataset_2", data="test")
    loaded_A = cebra_load.load(filename, key="dataset_1")
    return A, loaded_A


@register("h5", "hdf", "hdf5", "h")
def generate_h5_second(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    with h5py.File(filename, "w") as hf:
        hf.create_dataset("dataset_1", data="test")
        hf.create_dataset("dataset_2", data=A)
    loaded_A = cebra_load.load(filename)
    return A, loaded_A


@register_error("h5", "hdf", "hdf5", "h")
def generate_h5_wrong_key(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    with h5py.File(filename, "w") as hf:
        hf.create_dataset("dataset_1", data=A)
        hf.create_dataset("dataset_2", data="test")
    _ = cebra_load.load(filename, key="dataset_3")


@register_error("h5", "hdf", "hdf5", "h")
def generate_h5_invalid_key(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    with h5py.File(filename, "w") as hf:
        hf.create_dataset("dataset_1", data=A)
        hf.create_dataset("dataset_2", data="test")
    _ = cebra_load.load(filename, key="dataset_2")


@register_error("h5", "hdf", "hdf5", "h")
def generate_h5_no_array(filename, dtype):
    with h5py.File(filename, "w") as hf:
        hf.create_dataset("dataset_1", data="test_1")
        hf.create_dataset("dataset_2", data="test_2")
    _ = cebra_load.load(filename)


@register("h5", "hdf", "hdf5", "h")
def generate_h5_dataframe(filename, dtype):
    A = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    df_A = pd.DataFrame(np.array(A), columns=["a", "b", "c"])
    df_A.to_hdf(filename, key="df_A")
    loaded_A = cebra_load.load(filename, key="df_A")
    return A, loaded_A


@register("h5", "hdf", "hdf5", "h")
def generate_h5_dataframe_columns(filename, dtype):
    A = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    A_col = A[:, :2]
    df_A = pd.DataFrame(np.array(A), columns=["a", "b", "c"])
    df_A.to_hdf(filename, key="df_A")
    loaded_A = cebra_load.load(filename, key="df_A", columns=["a", "b"])
    return A_col, loaded_A


@register("h5", "hdf", "hdf5", "h")
def generate_h5_multi_dataframe(filename, dtype):
    A = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    B = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    df_A = pd.DataFrame(np.array(A), columns=["a", "b", "c"])
    df_B = pd.DataFrame(np.array(B), columns=["c", "d", "e"])
    df_A.to_hdf(filename, key="df_A")
    df_B.to_hdf(filename, key="df_B")
    loaded_A = cebra_load.load(filename, key="df_A")
    return A, loaded_A


@register("h5", "hdf", "hdf5", "h")
def generate_h5_single_dataframe_no_key(filename, dtype):
    A = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]).astype(dtype)
    df_A = pd.DataFrame(np.array(A), columns=["a", "b", "c"])
    df_A.to_hdf(filename, key="df_A")
    loaded_A = cebra_load.load(filename)
    return A, loaded_A


@register_error("h5", "hdf", "hdf5", "h")
def generate_h5_multi_dataframe_no_key(filename, dtype):
    A = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]).astype(dtype)
    B = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]).astype(dtype)
    df_A = pd.DataFrame(np.array(A), columns=["a", "b", "c"])
    df_B = pd.DataFrame(np.array(B), columns=["c", "d", "e"])
    df_A.to_hdf(filename, key="df_A")
    df_B.to_hdf(filename, key="df_B")
    _ = cebra_load.load(filename)


@register("h5", "hdf", "hdf5", "h")
def generate_h5_multicol_dataframe(filename, dtype):
    animals = ["mouse1", "mouse2"]
    keypoints = ["a", "b", "c"]
    data = [[[2, 4, 5], [3, 4, 5]], [[6, 7, 8], [9, 10, 11]]]
    A = np.array(data).reshape(2, len(keypoints) * len(animals)).astype(dtype)
    df_A = pd.DataFrame(A,
                        columns=pd.MultiIndex.from_product([animals,
                                                            keypoints]))
    df_A.to_hdf(filename, key="df_A")
    loaded_A = cebra_load.load(filename, key="df_A")
    return A, loaded_A


@register_error("h5", "hdf", "hdf5", "h")
def generate_h5_dataframe_invalid_key(filename, dtype):
    A = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]).astype(dtype)
    df_A = pd.DataFrame(np.array(A), columns=["a", "b", "c"])
    df_A.to_hdf(filename, key="df_A")
    _ = cebra_load.load(filename, key="df_B")


@register_error("h5", "hdf", "hdf5", "h")
def generate_h5_dataframe_invalid_column(filename, dtype):
    A = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]).astype(dtype)
    df_A = pd.DataFrame(np.array(A), columns=["a", "b", "c"])
    df_A.to_hdf(filename, key="df_A")
    _ = cebra_load.load(filename, key="df_A", columns=["d", "b"])


@register_error("h5", "hdf", "hdf5", "h")
def generate_h5_multicol_dataframe_columns(filename, dtype):
    animals = ["mouse1", "mouse2"]
    keypoints = ["a", "b", "c"]
    data = [[[2, 4, 5], [3, 4, 5]], [[6, 7, 8], [9, 10, 11]]]
    A = np.array(data).reshape(2, len(keypoints) * len(animals)).astype(dtype)
    df_A = pd.DataFrame(A,
                        columns=pd.MultiIndex.from_product([animals,
                                                            keypoints]))
    df_A.to_hdf(filename, key="df_A")
    _ = cebra_load.load(filename, key="df_A", columns=["a", "b"])


#### .PT ####
# TODO: test pytorch model is not loaded
# def test_load_torch_model():
#     assert False


@register("pt", "pth")
def generate_torch(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    A_tensor = torch.tensor(A)
    torch.save(A_tensor, filename)
    loaded_A = cebra_load.load(filename)
    return A, loaded_A


@register("pt", "pth")
def generate_torch_cofounder(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    A_tensor = torch.tensor(A)
    B = np.arange(500, dtype=dtype).reshape(10, 50)
    B_tensor = torch.tensor(B)
    torch.save({"A": A_tensor, "B": B_tensor}, filename)
    loaded_A = cebra_load.load(filename)
    return A, loaded_A


@register("pt", "pth")
def generate_torch_path(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    A_tensor = torch.tensor(A)
    B = np.arange(500, dtype=dtype).reshape(10, 50)
    B_tensor = torch.tensor(B)
    torch.save({"A": A_tensor, "B": B_tensor}, filename)
    loaded_A = cebra_load.load(pathlib.Path(filename))
    return A, loaded_A


@register("pt", "pth")
def generate_torch_key(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    A_tensor = torch.tensor(A)
    B = np.arange(500, dtype=dtype).reshape(10, 50)
    B_tensor = torch.tensor(B)
    torch.save({"A": A_tensor, "B": B_tensor}, filename)
    loaded_A = cebra_load.load(filename, key="A")
    return A, loaded_A


@register_error("pt", "pth")
def generate_wrong_key(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    A_tensor = torch.tensor(A)
    B = np.arange(500, dtype=dtype).reshape(10, 50)
    B_tensor = torch.tensor(B)
    torch.save({"A": A_tensor, "B": B_tensor}, filename)
    _ = cebra_load.load(filename, key="C")


#### .CSV ####
@register("csv", requires=("pandas",))
def generate_csv(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    pd.DataFrame(A).to_csv(filename, header=False, index=False, sep=",")
    loaded_A = cebra_load.load(filename)
    return A, loaded_A


@register("csv")
def generate_csv_path(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    pd.DataFrame(A).to_csv(filename, header=False, index=False, sep=",")
    loaded_A = cebra_load.load(pathlib.Path(filename))
    return A, loaded_A


@register_error("csv")
def generate_csv_empty_file(filename, dtype):
    with open(filename, "w") as _:
        pass
    _ = cebra_load.load(filename)


#### EXCEL ####
@register("xls", "xlsx", "xlsm", requires=("pandas", "pd"))
# TODO(celia): add the following extension:  "xlsb", "odf", "ods", "odt",
# issue to create the files
def generate_excel(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    A_df = pd.DataFrame(A)
    A_df.to_excel(filename, index=False, header=False)
    loaded_A = cebra_load.load(filename)
    return A, loaded_A


@register("xls", "xlsx", "xlsm")
def generate_excel_cofounder(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    A_df = pd.DataFrame(A)
    B = np.arange(500, dtype=dtype).reshape(10, 50)
    B_df = pd.DataFrame(B)
    with pd.ExcelWriter(filename) as writer:
        A_df.to_excel(writer, index=False, header=False, sheet_name="sheet_1")
        B_df.to_excel(writer, index=False, header=False, sheet_name="sheet_2")
    loaded_A = cebra_load.load(filename)
    return A, loaded_A


@register("xls", "xlsx", "xlsm")
def generate_excel_path(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    A_df = pd.DataFrame(A)
    B = np.arange(500, dtype=dtype).reshape(10, 50)
    B_df = pd.DataFrame(B)
    with pd.ExcelWriter(filename) as writer:
        A_df.to_excel(writer, index=False, header=False, sheet_name="sheet_1")
        B_df.to_excel(writer, index=False, header=False, sheet_name="sheet_2")
    loaded_A = cebra_load.load(pathlib.Path(filename))
    return A, loaded_A


@register("xls", "xlsx", "xlsm")
def generate_excel_key(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    A_df = pd.DataFrame(A)
    B = np.arange(500, dtype=dtype).reshape(10, 50)
    B_df = pd.DataFrame(B)
    with pd.ExcelWriter(filename) as writer:
        A_df.to_excel(writer, index=False, header=False, sheet_name="sheet_1")
        B_df.to_excel(writer, index=False, header=False, sheet_name="sheet_2")
    loaded_A = cebra_load.load(filename, key="sheet_1")
    return A, loaded_A


@register_error("xls", "xlsx", "xlsm")
def generate_excel_wrong_key(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    A_df = pd.DataFrame(A)
    B = np.arange(500, dtype=dtype).reshape(10, 50)
    B_df = pd.DataFrame(B)
    with pd.ExcelWriter(filename) as writer:
        A_df.to_excel(writer, index=False, header=False, sheet_name="sheet_1")
        B_df.to_excel(writer, index=False, header=False, sheet_name="sheet_2")
    _ = cebra_load.load(filename, key="sheet_3")


@register_error("xls", "xlsx", "xlsm")
def generate_excel_empty_file(filename, dtype):
    workbook = openpyxl.Workbook()
    workbook.save(filename)
    _ = cebra_load.load(filename)


#### .JL ####
@register("jl")
def generate_joblib(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    jl.dump(A, filename)
    loaded_A = cebra_load.load(filename)
    return A, loaded_A


@register("jl")
def generate_joblib_cofounder(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    print(filename)
    jl.dump({"A": A, "B": "test"}, filename)
    loaded_A = cebra_load.load(filename)
    return A, loaded_A


@register("jl")
def generate_joblib_path(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    jl.dump({"A": A, "B": "test"}, filename)
    loaded_A = cebra_load.load(pathlib.Path(filename))
    return A, loaded_A


@register("jl")
def generate_joblib_second(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    jl.dump({"B": "test", "A": A}, filename)
    loaded_A = cebra_load.load(filename)
    return A, loaded_A


@register("jl")
def generate_joblib_key(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    jl.dump({"A": A, "B": "test"}, filename)
    loaded_A = cebra_load.load(filename, key="A")
    return A, loaded_A


@register_error("jl")
def generate_joblib_wrong_key(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    jl.dump({"A": A, "B": "test"}, filename)
    _ = cebra_load.load(filename, key="C")


@register_error("jl")
def generate_joblib_invalid_key(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    jl.dump({"A": A, "B": "test"}, filename)
    _ = cebra_load.load(filename, key="B")


@register_error("jl")
def generate_joblib_no_array(filename, dtype):
    jl.dump({"A": "test_1", "B": "test_2"}, filename)
    _ = cebra_load.load(filename)


#### .PKL ####
@register("pkl", "p")
def generate_pickle(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    with open(filename, "wb") as f:
        pickle.dump(A, f)
    loaded_A = cebra_load.load(filename)
    return A, loaded_A


@register("pkl", "p")
def generate_pickle_cofounder(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    with open(filename, "wb") as f:
        pickle.dump({"A": A, "B": "test"}, f)
    loaded_A = cebra_load.load(filename)
    return A, loaded_A


@register("pkl", "p")
def generate_pickle_path(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    with open(filename, "wb") as f:
        pickle.dump({"A": A, "B": "test"}, f)
    loaded_A = cebra_load.load(pathlib.Path(filename))
    return A, loaded_A


@register("pkl", "p")
def generate_pickle_key(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    with open(filename, "wb") as f:
        pickle.dump({"A": A, "B": "test"}, f)
    loaded_A = cebra_load.load(filename, key="A")
    return A, loaded_A


@register("pkl", "p")
def generate_pickle_second(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    with open(filename, "wb") as f:
        pickle.dump({"B": "test", "A": A}, f)
    loaded_A = cebra_load.load(filename)
    return A, loaded_A


@register_error("pkl", "p")
def generate_pickle_wrong_key(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    with open(filename, "wb") as f:
        pickle.dump({"A": A, "B": "test"}, f)
    _ = cebra_load.load(filename, key="C")


@register_error("pkl", "p")
def generate_pickle_invalid_key(filename, dtype):
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    with open(filename, "wb") as f:
        pickle.dump({"A": A, "B": "test"}, f)
    _ = cebra_load.load(filename, key="B")


@register_error("pkl", "p")
def generate_pickle_no_array(filename, dtype):
    with open(filename, "wb") as f:
        pickle.dump({"A": "test_1", "B": "test_2"}, f)
    _ = cebra_load.load(filename)


#### DLC ####
# @register()
# def generate_load_dlc(filename):
#     assert False


#### .MAT ####
@register("mat")
def generate_mat_old(filename, dtype):
    """Older matplotlib arrays have their own format."""
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    A_dict = {"dataset_1": A, "label": "test"}
    scipy.io.savemat(filename, A_dict)
    loaded_A = cebra_load.load(filename)
    return A, loaded_A


@register("mat")
def generate_mat_old_path(filename, dtype):
    """Older matplotlib arrays have their own format."""
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    A_dict = {"dataset_1": A, "label": "test"}
    scipy.io.savemat(filename, A_dict)
    loaded_A = cebra_load.load(pathlib.Path(filename))
    return A, loaded_A


@register("mat")
def generate_mat_old_key(filename, dtype):
    """Older matplotlib arrays have their own format."""
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    A_dict = {"dataset_1": A, "label": "test"}
    scipy.io.savemat(filename, A_dict)
    loaded_A = cebra_load.load(filename, key="dataset_1")
    return A, loaded_A


@register("mat")
def generate_mat_old_second(filename, dtype):
    """Older matplotlib arrays have their own format."""
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    print(dtype)
    A_dict = {"label": "test", "dataset_1": A}
    scipy.io.savemat(filename, A_dict)
    loaded_A = cebra_load.load(filename)
    return A, loaded_A


@register_error("mat")
def generate_mat_old_wrong_key(filename, dtype):
    """Older matplotlib arrays have their own format."""
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    A_dict = {"dataset_1": A, "label": "test"}
    scipy.io.savemat(filename, A_dict)
    _ = cebra_load.load(filename, key="dataset_2")


@register_error("mat")
def generate_mat_old_invalid_key(filename, dtype):
    """Older matplotlib arrays have their own format."""
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    A_dict = {"dataset_1": A, "label": "test"}
    scipy.io.savemat(filename, A_dict)
    _ = cebra_load.load(filename, key="label")


@register_error("mat")
def generate_mat_old_no_array(filename, dtype):
    """Older matplotlib arrays have their own format."""
    A_dict = {"dataset_1": "test_1", "dataset_2": "test_2"}
    scipy.io.savemat(filename, A_dict)
    _ = cebra_load.load(filename)


@register("mat")
def generate_mat_new(filename, dtype):
    """Newer matplotlib formats are just h5 files"""
    _skip_hdf5storage()
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    # A transposed as matrices are transposed to be stored in HDF5
    A_dict = {"dataset_1": A.T, "label": "test"}
    hdf5storage.write(A_dict, ".", filename, matlab_compatible=True)
    loaded_A = cebra_load.load(filename)
    return A, loaded_A


@register("mat")
def generate_mat_new_path(filename, dtype):
    """Newer matplotlib formats are just h5 files"""
    _skip_hdf5storage()
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    # A transposed as matrices are transposed to be stored in HDF5
    A_dict = {"dataset_1": A.T, "label": "test"}
    hdf5storage.write(A_dict, ".", filename, matlab_compatible=True)
    loaded_A = cebra_load.load(pathlib.Path(filename))
    return A, loaded_A


@register("mat")
def generate_mat_new_key(filename, dtype):
    """Newer matplotlib formats are just h5 files"""
    _skip_hdf5storage()
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    # A transposed as matrices are transposed to be stored in HDF5
    A_dict = {"dataset_1": A.T, "label": "test"}
    hdf5storage.write(A_dict, ".", filename, matlab_compatible=True)
    loaded_A = cebra_load.load(filename, key="dataset_1")
    return A, loaded_A


@register("mat")
def generate_mat_new_second(filename, dtype):
    """Newer matplotlib formats are just h5 files"""
    _skip_hdf5storage()
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    # A transposed as matrices are transposed to be stored in HDF5
    A_dict = {"label": "test", "dataset_1": A.T}
    hdf5storage.write(A_dict, ".", filename, matlab_compatible=True)
    loaded_A = cebra_load.load(filename)
    return A, loaded_A


@register_error("mat")
def generate_mat_new_wrong_key(filename, dtype):
    """Newer matplotlib formats are just h5 files"""
    _skip_hdf5storage()
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    # A transposed as matrices are transposed to be stored in HDF5
    A_dict = {"dataset_1": A.T, "label": "test"}
    hdf5storage.write(A_dict, ".", filename, matlab_compatible=True)
    _ = cebra_load.load(filename, key="dataset_2")


@register_error("mat")
def generate_mat_new_invalid_key(filename, dtype):
    """Newer matplotlib formats are just h5 files"""
    _skip_hdf5storage()
    A = np.arange(1000, dtype=dtype).reshape(10, 100)
    # A transposed as matrices are transposed to be stored in HDF5
    A_dict = {"dataset_1": A.T, "label": "test"}
    hdf5storage.write(A_dict, ".", filename, matlab_compatible=True)
    _ = cebra_load.load(filename, key="label")


@register_error("mat")
def generate_mat_new_no_array(filename, dtype):
    """Newer matplotlib formats are just h5 files"""
    _skip_hdf5storage()
    A_dict = {"dataset_1": "test_1", "label": "test_2"}
    hdf5storage.write(A_dict, ".", filename, matlab_compatible=True)
    _ = cebra_load.load(filename)


test_functions = list(itertools.product(__test_functions, [np.int32, np.int64]))
test_functions = [(*t, x) for t, x in test_functions]


@pytest.mark.parametrize("save_data,file_ending,dtype", test_functions)
def test_load(save_data, file_ending, dtype):
    with tempfile.NamedTemporaryFile() as tf:
        filename = tf.name  # name, without extension

    if file_ending in ("csv", "xls", "xlsx", "xlsm"):
        if dtype == np.int32:
            pytest.skip(
                "Skipping test. For CSV, XLS, XLSX, and XLM file formats, "
                "the integer loading data type is always int64, regardless of the "
                "data type it was saved with. This can lead to compatibility issues, "
                "especially on Windows. To ensure accurate testing, we only perform "
                "tests with int64 data type for these formats, and we skip the test "
                "cases involving int32.")

    # create data, save it, load it
    saved_array, loaded_array = save_data(filename, dtype=dtype)

    assert isinstance(loaded_array, np.ndarray)
    assert loaded_array.dtype == saved_array.dtype
    assert loaded_array.shape == saved_array.shape
    assert np.allclose(saved_array, loaded_array)


@pytest.mark.parametrize("save_data", __test_functions_error)
def test_load_error(save_data):
    with tempfile.NamedTemporaryFile() as tf:
        filename = tf.name  # name, without extension

    with pytest.raises((AttributeError, TypeError)):
        save_data(filename)


test_functions_module_not_found = list(
    itertools.product(__test_functions_module_not_found, [np.int32, np.int64]))
test_functions_module_not_found = [
    (*t, x) for t, x in test_functions_module_not_found
]


@pytest.mark.parametrize("module_names,get_path,save_data,dtype",
                         test_functions_module_not_found)
def test_module_not_installed(module_names, get_path, save_data, dtype):

    assert len(module_names) > 0
    assert isinstance(module_names, tuple)

    with tempfile.NamedTemporaryFile() as tf:
        filename = tf.name

    saved_array, loaded_array = save_data(filename, dtype=dtype)
    assert np.allclose(saved_array, loaded_array)

    # TODO(stes): Sketch for a test --- needs additional work.
    # with patch.dict('sys.modules', {module: None for module in module_names}):
    #    path = get_path(filename)
    #    with pytest.raises(ModuleNotFoundError, match="cebra[datasets]"):
    #        cebra.data.load.load(path)



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np
import pytest
import torch

import cebra.helper
import cebra.integrations.sklearn.decoder as cebra_sklearn_decoder


def test_imports():
    import cebra

    assert hasattr(cebra, "KNNDecoder")
    assert hasattr(cebra, "L1LinearRegressor")


def _iterate_decoders():
    decoders = []
    for param in cebra_sklearn_decoder.KNNDecoder.iter_hyperparams():
        decoders.append(
            cebra_sklearn_decoder.KNNDecoder(n_neighbors=param["n_neighbors"],
                                             metric=param["metric"]))
    for param in cebra_sklearn_decoder.L1LinearRegressor.iter_hyperparams():
        decoders.append(
            cebra_sklearn_decoder.L1LinearRegressor(alpha=param["alpha"]))
    return decoders


@pytest.mark.parametrize("decoder", _iterate_decoders())
def test_sklearn_decoder(decoder):
    # example dataset
    X = np.random.uniform(0, 1, (1000, 50))
    y_c = np.random.uniform(0, 1, (1000))
    y_d = np.random.randint(0, 10, (1000,))
    y_c_dim = np.random.uniform(0, 1, (1000, 5))
    y_str = np.array(["test" for i in range(1000)])
    y_d_short = np.random.randint(0, 10, (500,))

    # continuous target
    decoder.fit(X, y_c)
    pred = decoder.predict(X)
    assert isinstance(pred, np.ndarray)
    assert np.issubdtype(pred.dtype, np.floating)

    score = decoder.score(X, y_c)
    assert isinstance(score, float)

    # torch
    decoder.fit(torch.Tensor(X), torch.Tensor(y_c))
    pred = decoder.predict(torch.Tensor(X))
    assert isinstance(pred, np.ndarray)
    assert np.issubdtype(pred.dtype, np.floating)

    # discrete target
    decoder.fit(X, y_d)
    pred = decoder.predict(X)
    assert isinstance(pred, np.ndarray)
    assert np.issubdtype(pred.dtype, np.integer) or np.issubdtype(
        pred.dtype, np.floating)

    score = decoder.score(X, y_d)
    assert isinstance(score, float)

    # torch
    decoder.fit(torch.Tensor(X), torch.Tensor(y_d))
    pred = decoder.predict(torch.Tensor(X))
    assert isinstance(pred, np.ndarray)
    assert np.issubdtype(pred.dtype, np.integer) or np.issubdtype(
        pred.dtype, np.floating)

    # multi-dim continuous target
    decoder.fit(X, y_c_dim)
    pred = decoder.predict(X)
    assert isinstance(pred, np.ndarray)
    assert np.issubdtype(pred.dtype, np.floating)

    score = decoder.score(X, y_c_dim)
    assert isinstance(score, float)

    # multi-dim discrete and continuous target
    multi_y = np.concatenate([y_c_dim, np.expand_dims(y_d, axis=1)], axis=1)
    decoder.fit(X, multi_y)
    pred = decoder.predict(X)
    assert isinstance(pred, np.ndarray)
    assert np.issubdtype(pred.dtype, np.floating)

    score = decoder.score(X, multi_y)
    assert isinstance(score, float)

    # invalid targets
    with pytest.raises(NotImplementedError, match="Invalid.*type"):
        decoder.fit(X, y_str)
    with pytest.raises(ValueError, match="Invalid.*shape"):
        decoder.fit(X, y_d_short)


def test_dtype_checker():
    assert cebra.helper._is_floating(torch.Tensor([4.5]))
    assert cebra.helper._is_integer(torch.LongTensor([4]))
    assert cebra.helper._is_floating(np.array([4.5]))
    assert cebra.helper._is_integer(np.array([4]))



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import copy
import itertools

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pkg_resources
import pytest
import torch
from sklearn.exceptions import NotFittedError

import cebra.integrations.matplotlib as cebra_plot
import cebra.integrations.sklearn.cebra as cebra_sklearn_cebra
import cebra.integrations.sklearn.metrics as cebra_sklearn_metrics


class _ConsistencyDataset():
    """A test dataset that can be indexed to obtain different permutations"""

    def __init__(self, seed=42424242):
        self._generator = np.random.default_rng(seed=seed)
        self.dataset_ids = "a", "b", "c"
        self.embeddings = [
            self._generator.normal(size=(1000, 3)) for _ in self.dataset_ids
        ]
        self.labels = [
            self._generator.uniform(0, 1, size=(1000,))
            for _ in self.dataset_ids
        ]

    def __getitem__(self, order):
        return {
            "dataset_ids": [self.dataset_ids[i] for i in order],
            "embeddings": [self.embeddings[i] for i in order],
            "labels": [self.labels[i] for i in order]
        }


def test_plot_imports():
    import cebra

    assert hasattr(cebra, "plot")
    assert hasattr(cebra, "plot_embedding")
    assert hasattr(cebra, "plot_temperature")
    assert hasattr(cebra, "plot_loss")
    assert hasattr(cebra, "plot_overview")
    assert hasattr(cebra, "compare_models")
    assert hasattr(cebra, "plot_consistency")


def test_colormaps():
    import matplotlib

    cmap = matplotlib.colormaps["cebra"]
    assert cmap is not None
    plt.scatter([1], [2], c=[2], cmap="cebra")


def test_plot_overview():
    # example dataset
    X = np.random.uniform(0, 1, (1000, 50))
    y_c1 = np.random.uniform(0, 1, (1000, 5))
    y_c2 = np.random.uniform(0, 1, (800, 2))
    y_d = np.random.randint(0, 10, (1000,))

    # define a simple CEBRA model
    model = cebra_sklearn_cebra.CEBRA(max_iterations=10, batch_size=512)

    with pytest.raises(NotFittedError, match="not.*fitted"):
        _ = cebra_plot.plot_overview(model, X)
        plt.close()

    model.fit(X)

    fig, (ax1, ax2, ax3) = cebra_plot.plot_overview(model, X)
    assert isinstance(fig, matplotlib.figure.Figure)
    assert isinstance(ax1, matplotlib.axes.Axes)
    assert isinstance(ax2, matplotlib.axes.Axes)
    assert isinstance(ax3, matplotlib.axes.Axes)
    plt.close()
    fig, (ax1, ax2, ax3) = cebra_plot.plot_overview(
        model, X, embedding_kwargs={"embedding_labels": y_c1[:, 0]})
    assert isinstance(fig, matplotlib.figure.Figure)
    assert isinstance(ax1, matplotlib.axes.Axes)
    assert isinstance(ax2, matplotlib.axes.Axes)
    assert isinstance(ax3, matplotlib.axes.Axes)
    plt.close()
    fig, (ax1, ax2, ax3) = cebra_plot.plot_overview(
        model, X, embedding_kwargs={"embedding_labels": y_d})
    assert isinstance(fig, matplotlib.figure.Figure)
    assert isinstance(ax1, matplotlib.axes.Axes)
    assert isinstance(ax2, matplotlib.axes.Axes)
    assert isinstance(ax3, matplotlib.axes.Axes)
    plt.close()

    with pytest.raises(ValueError, match="Invalid.*embedding.*labels"):
        _ = cebra_plot.plot_overview(
            model, X, embedding_kwargs={"embedding_labels": y_c2[:, 0]})
        plt.close()
    with pytest.raises(NotImplementedError, match="Invalid.*embedding.*labels"):
        _ = cebra_plot.plot_overview(
            model, X, embedding_kwargs={"embedding_labels": y_c1[:, :1]})
        plt.close()


def test_plot_temperature():
    # example dataset
    X = np.random.uniform(0, 1, (1000, 50))

    fig = plt.figure(figsize=(5, 5))
    ax = fig.add_subplot()

    # define a simple CEBRA model
    model = cebra_sklearn_cebra.CEBRA(max_iterations=10, batch_size=512)

    with pytest.raises(NotFittedError, match="not.*fitted"):
        _ = cebra_plot.plot_temperature(model, ax=ax)
    with pytest.raises(NotFittedError, match="not.*fitted"):
        _ = cebra_plot.plot(model, which="temperature", ax=ax)

    model.fit(X)

    ax = cebra_plot.plot_temperature(model, ax=ax)
    assert isinstance(ax, matplotlib.axes.Axes)
    ax = cebra_plot.plot(model, which="temperature", ax=ax)
    assert isinstance(ax, matplotlib.axes.Axes)

    plt.close()


def test_plot_loss():
    # example dataset
    X = np.random.uniform(0, 1, (1000, 50))

    fig = plt.figure(figsize=(5, 5))
    ax = fig.add_subplot()

    # define a simple CEBRA model
    model = cebra_sklearn_cebra.CEBRA(max_iterations=10, batch_size=512)

    with pytest.raises(NotFittedError, match="not.*fitted"):
        _ = cebra_plot.plot_loss(model, ax=ax)
    with pytest.raises(NotFittedError, match="not.*fitted"):
        _ = cebra_plot.plot(model, which="loss", ax=ax)

    model.fit(X)

    ax = cebra_plot.plot_loss(model, ax=ax)
    assert isinstance(ax, matplotlib.axes.Axes)
    ax = cebra_plot.plot(model, which="loss", ax=ax)
    assert isinstance(ax, matplotlib.axes.Axes)
    ax = cebra_plot.plot(model, which="loss", ax=ax, label="test")
    assert isinstance(ax, matplotlib.axes.Axes)
    plt.close()


@pytest.mark.parametrize("matplotlib_version",
                         ["3.3", "3.4.2", "3.5", "3.6", "3.7"])
def test_compare_models_with_different_versions(matplotlib_version):
    # example dataset
    X = np.random.uniform(0, 1, (1000, 2))
    n_models = 2

    fitted_models = []
    for _ in range(n_models):
        fitted_models.append(
            cebra_sklearn_cebra.CEBRA(max_iterations=10, batch_size=128).fit(X))

    # minimum version of matplotlib
    minimum_version = "3.6"

    if pkg_resources.parse_version(
            matplotlib_version) < pkg_resources.parse_version(minimum_version):
        with pytest.raises(ImportError):
            cebra_plot.compare_models(models=fitted_models,
                                      patched_version=matplotlib_version)


def test_compare_models():
    # example dataset
    X = np.random.uniform(0, 1, (100, 5))
    n_models = 4

    fig = plt.figure(figsize=(5, 5))
    ax = fig.add_subplot()

    models, labels = [], []
    # define CEBRA models
    for i in range(n_models):
        models.append(
            cebra_sklearn_cebra.CEBRA(max_iterations=10, batch_size=512))
        labels.append(f"model_{i}")

    with pytest.raises(NotFittedError, match="not.*fitted"):
        _ = cebra_plot.compare_models(models, ax=ax)

    for model in models:
        model.fit(X)

    ax = cebra_plot.compare_models(models, ax=ax)
    assert isinstance(ax, matplotlib.axes.Axes)
    ax = cebra_plot.compare_models(models, ax=ax, cmap="viridis")
    assert isinstance(ax, matplotlib.axes.Axes)
    ax = cebra_plot.compare_models(models, ax=ax, labels=labels)
    assert isinstance(ax, matplotlib.axes.Axes)

    with pytest.raises(ValueError, match="Invalid.*models"):
        _ = cebra_plot.compare_models("test", ax=ax)
    with pytest.raises(ValueError, match="Invalid.*models"):
        invalid_models = copy.deepcopy(models)
        invalid_models.append("test")
        _ = cebra_plot.compare_models(invalid_models, ax=ax)
    with pytest.raises(ValueError, match="Invalid.*labels"):
        _ = cebra_plot.compare_models(models, labels="test", ax=ax)
    with pytest.raises(ValueError, match="Invalid.*labels"):
        long_labels = copy.deepcopy(labels)
        long_labels.append("test")
        _ = cebra_plot.compare_models(models, labels=long_labels, ax=ax)
    with pytest.raises(ValueError, match="Invalid.*labels"):
        invalid_labels = copy.deepcopy(labels)
        _ = invalid_labels.pop()
        invalid_labels.append(["a"])
        _ = cebra_plot.compare_models(models, labels=invalid_labels, ax=ax)

    plt.close()


@pytest.mark.parametrize("output_dimension, idx_order", [(8, (2, 3, 4)),
                                                         (2, (0, 1))])
def test_plot_embedding(output_dimension, idx_order):
    # example dataset
    X = np.random.uniform(0, 1, (1000, 50))
    y_c1 = np.random.uniform(0, 1, (1000, 5))
    y_c2 = np.random.uniform(0, 1, (800, 2))
    y_d = np.random.randint(0, 10, (1000,))

    fig = plt.figure(figsize=(5, 5))
    if output_dimension < 3:
        ax = fig.add_subplot()
    else:
        ax = fig.add_subplot(projection="3d")

    # integration tests
    model = cebra_sklearn_cebra.CEBRA(max_iterations=10,
                                      batch_size=512,
                                      output_dimension=output_dimension)

    with pytest.raises(NotFittedError, match="not.*fitted"):
        _ = cebra_plot.plot(model, which="embedding", ax=ax, X=X)

    model.fit(X)
    embedding = model.transform(X)

    ax = cebra_plot.plot_embedding(embedding=embedding)
    assert isinstance(ax, matplotlib.axes.Axes)
    ax = cebra_plot.plot_embedding(embedding=embedding, ax=ax)
    assert isinstance(ax, matplotlib.axes.Axes)
    ax = cebra_plot.plot_embedding(embedding=embedding,
                                   embedding_labels=y_c1[:, 0],
                                   ax=ax)
    assert isinstance(ax, matplotlib.axes.Axes)
    ax = cebra_plot.plot_embedding(embedding=embedding,
                                   idx_order=idx_order,
                                   ax=ax)
    assert isinstance(ax, matplotlib.axes.Axes)
    ax = cebra_plot.plot_embedding(embedding=embedding,
                                   embedding_labels=y_d,
                                   ax=ax)
    assert isinstance(ax, matplotlib.axes.Axes)

    plt.close()

    with pytest.raises(ValueError):
        _ = cebra_plot.plot_embedding(embedding=embedding,
                                      idx_order=(10, 13, 15),
                                      ax=ax)
    with pytest.raises(ValueError):
        _ = cebra_plot.plot_embedding(embedding=embedding,
                                      idx_order=(10, 13),
                                      ax=ax)
    with pytest.raises(ValueError, match="Invalid.*embedding.*labels"):
        _ = cebra_plot.plot_embedding(embedding,
                                      embedding_labels=y_c2[:, 0],
                                      ax=ax)
    with pytest.raises(NotImplementedError, match="Invalid.*embedding.*labels"):
        _ = cebra_plot.plot_embedding(embedding,
                                      embedding_labels=y_c1[:, :1],
                                      ax=ax)

    ax = cebra_plot.plot(model, which="embedding", X=X, ax=ax)
    assert isinstance(ax, matplotlib.axes.Axes)
    ax = cebra_plot.plot(model,
                         which="embedding",
                         X=X,
                         embedding_labels=y_c1[:, 0],
                         ax=ax)
    assert isinstance(ax, matplotlib.axes.Axes)
    ax = cebra_plot.plot(model,
                         which="embedding",
                         X=X,
                         embedding_labels=y_c1[:, 0],
                         ax=ax)

    with pytest.raises(ValueError, match="Invalid.*embedding.*labels"):
        _ = cebra_plot.plot(model,
                            which="embedding",
                            X=X,
                            embedding_labels=y_c2[:, 0],
                            ax=ax)
    with pytest.raises(NotImplementedError, match="Invalid.*embedding.*labels"):
        _ = cebra_plot.plot(model,
                            which="embedding",
                            X=X,
                            embedding_labels=y_c1[:, :1],
                            ax=ax)

    ax = cebra_plot.plot_embedding(embedding=embedding, idx_order=(0, 1))
    assert isinstance(ax, matplotlib.axes.Axes)
    ax = cebra_plot.plot_embedding(embedding=embedding, idx_order=(0, 1, 1))
    assert isinstance(ax, matplotlib.axes.Axes)

    plt.close()


def test_plot_consistency():
    embedding1 = np.random.uniform(0, 1, (1000, 4))
    embedding2 = np.random.uniform(0, 1, (1000, 10))
    embedding3 = np.random.uniform(0, 1, (800, 6))
    embedding4 = np.random.uniform(0, 1, (500, 7))
    embeddings_datasets = [embedding1, embedding2, embedding3, embedding4]
    embeddings_runs = [embedding1, embedding2, embedding1, embedding2]

    labels1 = np.random.uniform(0, 1, (1000,))
    labels2 = np.random.uniform(0, 1, (1000,))
    labels3 = np.random.uniform(0, 1, (800,))
    labels4 = np.random.uniform(0, 1, (500,))
    labels_datasets = [labels1, labels2, labels3, labels4]

    dataset_ids = ["achilles", "buddy", "cicero", "gatsby"]

    figure = plt.figure(figsize=(5, 5))
    ax = figure.add_subplot()

    scores_subs, pairs_subs, datasets_subs = cebra_sklearn_metrics.consistency_score(
        embeddings_datasets,
        labels=labels_datasets,
        dataset_ids=dataset_ids,
        between="datasets",
    )
    scores_runs, pairs_runs, datasets_runs = cebra_sklearn_metrics.consistency_score(
        embeddings_runs, between="runs")

    # between datasets
    fig = cebra_plot.plot_consistency(scores_subs,
                                      pairs=pairs_subs,
                                      datasets=datasets_subs)
    assert isinstance(fig, matplotlib.axes.Axes)
    plt.close()
    ax = cebra_plot.plot_consistency(scores_subs,
                                     pairs=pairs_subs,
                                     datasets=datasets_subs,
                                     ax=ax)
    assert isinstance(ax, matplotlib.axes.Axes)
    ax = cebra_plot.plot_consistency(
        torch.from_numpy(scores_subs),
        pairs=pairs_subs,
        datasets=datasets_subs,
        cmap="viridis",
        title="Test",
        text_color=None,
        colorbar_label=None,
        ax=ax,
    )
    assert isinstance(fig, matplotlib.axes.Axes)

    ax = cebra_plot.plot_consistency(torch.from_numpy(scores_subs),
                                     pairs=pairs_subs,
                                     datasets=datasets_subs,
                                     ax=ax)
    assert isinstance(fig, matplotlib.axes.Axes)

    ax = cebra_plot.plot_consistency(
        scores_subs.tolist(),
        pairs=pairs_subs.tolist(),
        datasets=datasets_subs.tolist(),
        ax=ax,
    )
    assert isinstance(fig, matplotlib.axes.Axes)

    with pytest.raises(ValueError, match="Missing.*datasets.*pairs"):
        _ = cebra_plot.plot_consistency(scores_subs, ax=ax)
    with pytest.raises(ValueError, match="Missing.*datasets.*pairs"):
        _ = cebra_plot.plot_consistency(scores_subs, pairs=pairs_subs, ax=ax)
    with pytest.raises(ValueError, match="Missing.*datasets.*pairs"):
        _ = cebra_plot.plot_consistency(scores_subs,
                                        datasets=datasets_subs,
                                        ax=ax)
    with pytest.raises(ValueError, match="Shape.*pairs"):
        _ = cebra_plot.plot_consistency(
            scores_subs,
            pairs=np.random.uniform(0, 1, (10, 2)),
            datasets=datasets_subs,
            ax=ax,
        )
    with pytest.raises(ValueError, match="Shape.*datasets"):
        _ = cebra_plot.plot_consistency(
            scores_subs,
            pairs=np.random.uniform(0, 1, (10, 2)),
            datasets=np.random.uniform(0, 1, (2,)),
            ax=ax,
        )
    with pytest.raises(ValueError, match="Invalid.*scores"):
        _ = cebra_plot.plot_consistency(
            np.random.uniform(0, 1, (12, 2, 2)),
            pairs=pairs_subs,
            datasets=datasets_subs,
            ax=ax,
        )

    # between runs
    fig = cebra_plot.plot_consistency(scores_runs,
                                      pairs=pairs_runs,
                                      datasets=datasets_runs)
    assert isinstance(fig, matplotlib.axes.Axes)
    plt.close()
    ax = cebra_plot.plot_consistency(scores_runs,
                                     pairs=pairs_runs,
                                     datasets=datasets_runs,
                                     ax=ax)
    assert isinstance(ax, matplotlib.axes.Axes)
    ax = cebra_plot.plot_consistency(
        scores_runs,
        pairs=pairs_runs,
        datasets=datasets_runs,
        cmap="viridis",
        title="Test",
        text_color=None,
        colorbar_label=None,
        ax=ax,
    )
    assert isinstance(ax, matplotlib.axes.Axes)
    ax = cebra_plot.plot_consistency(torch.from_numpy(scores_runs),
                                     pairs=pairs_runs,
                                     datasets=datasets_runs,
                                     ax=ax)
    assert isinstance(ax, matplotlib.axes.Axes)
    ax = cebra_plot.plot_consistency(
        scores_runs.tolist(),
        pairs=pairs_runs.tolist(),
        datasets=datasets_runs.tolist(),
        ax=ax,
    )
    assert isinstance(ax, matplotlib.axes.Axes)

    with pytest.raises(ValueError, match="Missing.*datasets.*pairs"):
        _ = cebra_plot.plot_consistency(scores_runs, ax=ax)
    with pytest.raises(ValueError, match="Missing.*datasets.*pairs"):
        _ = cebra_plot.plot_consistency(scores_runs, pairs=pairs_runs, ax=ax)
    with pytest.raises(ValueError, match="Missing.*datasets.*pairs"):
        _ = cebra_plot.plot_consistency(scores_runs,
                                        datasets=datasets_runs,
                                        ax=ax)
    with pytest.raises(ValueError, match="Shape.*datasets"):
        _ = cebra_plot.plot_consistency(
            scores_runs,
            pairs=np.random.uniform(0, 1, (10, 2)),
            datasets=np.random.uniform(0, 1, (4,)),
            ax=ax,
        )
    with pytest.raises(ValueError, match="Shape.*pairs"):
        _ = cebra_plot.plot_consistency(
            scores_runs,
            pairs=np.random.uniform(0, 1, (10, 2)),
            datasets=datasets_runs,
            ax=ax,
        )
    with pytest.raises(ValueError, match="Invalid.*dimensions"):
        _ = cebra_plot.plot_consistency(
            np.random.uniform(0, 1, (12, 2, 2)),
            pairs=pairs_runs,
            datasets=datasets_runs,
            ax=ax,
        )
    plt.close()


@pytest.mark.parametrize("seed", [None, 42, 1024, 454545])
def test_check_consistency_data(seed):
    """check the helper functions that generates the dataset for tests below."""
    if seed is None:
        data1 = _ConsistencyDataset()
        data2 = _ConsistencyDataset()
    else:
        data1 = _ConsistencyDataset(seed=seed)
        data2 = _ConsistencyDataset(seed=seed)

    assert all(
        np.allclose(a, b) for a, b in zip(data1.embeddings, data2.embeddings))
    assert all(np.allclose(a, b) for a, b in zip(data1.labels, data2.labels))
    assert all(a == b for a, b in zip(data1.dataset_ids, data2.dataset_ids))


@pytest.mark.parametrize("permutation", itertools.permutations([0, 1, 2]))
def test_consistency_permutations(permutation):
    dataset = _ConsistencyDataset()

    def compute_consistency(kwargs):
        scores, pairs, datasets = cebra_sklearn_metrics.consistency_score(
            between="datasets", **kwargs)
        return {tuple(pair): score for pair, score in zip(pairs, scores)}

    assert compute_consistency(dataset[0, 1,
                                       2]) != compute_consistency(dataset[1, 1,
                                                                          0])
    assert compute_consistency(dataset[0, 1, 2]) == compute_consistency(
        dataset[permutation])


@pytest.mark.parametrize("seed,permutation,pass_original_labels",
                         [(42, (0, 1, 2), False), (43, (0, 2, 1), False),
                          (44, (0, 2, 1), True), (45, (1, 2, 0), False)])
def test_plot_consistency_unordered_labels(seed, permutation,
                                           pass_original_labels):

    dataset = _ConsistencyDataset(seed=seed)

    scores, pairs, datasets = cebra_sklearn_metrics.consistency_score(
        between="datasets", **dataset[permutation])
    scores_dict = {tuple(pair): score for pair, score in zip(pairs, scores)}

    plot = cebra_plot._ConsistencyPlot(
        scores=scores,
        pairs=pairs,
        datasets=dataset[permutation]["dataset_ids"]
        if pass_original_labels else datasets,
        cmap="cebra",
        text_color="black",
        colorbar_label=None,
        title="testing",
        axis=plt.gca(),
        figsize=(3, 3),
        dpi=200)

    for i, label_i in enumerate(plot.labels):
        for j, label_j in enumerate(plot.labels):
            if i == j:
                # these values are nan, just check for that
                assert np.isnan(plot.scores[i, j])
            else:
                assert np.isclose(100 * scores_dict[label_i, label_j],
                                  plot.scores[i, j])



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Allen single mouse dataset.

References:
    * Deitch, Daniel, Alon Rubin, and Yaniv Ziv.
      "Representational drift in the mouse visual cortex."
      Current Biology 31.19 (2021): 4327-4339.

    * de Vries, Saskia EJ, et al.
      "A large-scale standardized physiological survey reveals functional
      organization of the mouse visual cortex."
      Nature Neuroscience 23.1 (2020): 138-151.

    * https://github.com/zivlab/visual_drift
    * http://observatory.brain-map.org/visualcoding
"""
import pathlib

import numpy as np
import scipy.io
import torch
from sklearn.decomposition import PCA

import cebra.data
from cebra.datasets import get_datapath
from cebra.datasets import init
from cebra.datasets import parametrize

_DEFAULT_DATADIR = get_datapath()

_SINGLE_SESSION_CA = (
    pathlib.Path(_DEFAULT_DATADIR) / "allen" / "visual_drift" / "data" /
    "calcium_excitatory" / "VISp" / "680156909.mat",
    pathlib.Path(_DEFAULT_DATADIR) / "allen" / "visual_drift" / "data" /
    "calcium_excitatory" / "VISp" / "511510779.mat",
    pathlib.Path(_DEFAULT_DATADIR) / "allen" / "visual_drift" / "data" /
    "calcium_excitatory" / "VISp" / "679702882.mat",
    pathlib.Path(_DEFAULT_DATADIR) / "allen" / "visual_drift" / "data" /
    "calcium_excitatory" / "VISp" / "688678764.mat",
)


@parametrize(
    "allen-movie1-ca-single-session-{session_id}",
    session_id=range(len(_SINGLE_SESSION_CA)),
)
class SingleSessionAllenCa(cebra.data.SingleSessionDataset):
    """A single mouse 30Hz calcium events dataset during the allen MOVIE1 stimulus.

    A dataset of a single mouse 30Hz calcium events from the excitatory neurons in the primary visual cortex
    during the 10 repeats of the MOVIE1 stimulus in session type A. The preprocessed data from *Deitch et al. (2021) are used.
    The continuous labels corresponding to a DINO embedding of each stimulus frame.

    Args:
        session_id: The integer value to pick a session among 4 sessions with the largest number of recorded neruons. Choose between 0-3.
        frame_feature_path: The path of the movie frame features.

    """

    def __init__(
        self,
        session_id: int,
        frame_feature_path: str = pathlib.Path(_DEFAULT_DATADIR) / "allen" /
        "features" / "allen_movies" / "vit_base" / "8" /
        "movie_one_image_stack.npz" / "testfeat.pth",
        pca: bool = False,
    ):
        self.path = _SINGLE_SESSION_CA[session_id]
        traces = scipy.io.loadmat(self.path)
        if pca:
            pca_ = PCA()
            neural = pca_.fit_transform(
                traces["filtered_traces_days_events"][0,
                                                      0].transpose(1,
                                                                   0))[:, :32]
        else:
            neural = traces["filtered_traces_days_events"][0, 0].transpose(1, 0)
        self.neural = torch.from_numpy(neural).float()
        frame_feature = torch.load(frame_feature_path)
        self.index = frame_feature.repeat(10, 1)

    def __len__(self):
        return self.neural.size(0)

    @property
    def continuous_index(self):
        return self.index

    @property
    def input_dimension(self):
        return self.neural.size(1)

    def __getitem__(self, index):
        index = self.expand_index(index)
        return self.neural[index].transpose(2, 1)


@parametrize(
    "allen-movie1-ca-single-session-corrupt-{session_id}",
    session_id=range(len(_SINGLE_SESSION_CA)),
)
class SingleSessionAllenCaCorrupted(cebra.data.SingleSessionDataset):
    """A corrupted single mouse 30Hz calcium events dataset during the allen MOVIE1 stimulus.

    A dataset of a single mouse 30Hz calcium events from the excitatory neurons in the primary visual cortex
    during the 10 repeats of the MOVIE1 stimulus in session type A. The preprocessed data from *Deitch et al. (2021) are used.
    The continuous labels corresponding to a DINO embedding of each stimulus frame, but in randomly shuffled order.

    Args:
        session_id: The integer value to pick a session among 4 sessions with the largest number of recorded neruons. Choose between 0-3.
        frame_feature_path: The path of the movie frame features.

    """

    def __init__(
        self,
        session_id: int,
        frame_feature_path: str = pathlib.Path(_DEFAULT_DATADIR) / "allen" /
        "features" / "allen_movies" / "vit_base" / "8" /
        "movie_one_image_stack.npz" / "testfeat.pth",
        pca: bool = False,
    ):
        self.path = _SINGLE_SESSION_CA[session_id]
        traces = scipy.io.loadmat(self.path)
        if pca:
            pca_ = PCA()
            neural = pca_.fit_transform(
                traces["filtered_traces_days_events"][0,
                                                      0].transpose(1,
                                                                   0))[:, :32]
        else:
            neural = traces["filtered_traces_days_events"][0, 0].transpose(1, 0)

        self.neural = torch.from_numpy(neural).float()
        frame_feature = torch.load(frame_feature_path)
        self.frame_index = np.arange(900)
        rng = np.random.Generator(np.random.PCG64(111))
        rng.shuffle(self.frame_index)
        self.index = frame_feature.repeat(10, 1)[np.tile(self.frame_index, 10)]

    def __len__(self):
        return self.neural.size(0)

    @property
    def continuous_index(self):
        return self.index

    @property
    def input_dimension(self):
        return self.neural.size(1)

    def __getitem__(self, index):
        index = self.expand_index(index)
        return self.neural[index].transpose(2, 1)


@parametrize(
    "allen-movie1-ca-single-session-time-{session_id}",
    session_id=range(len(_SINGLE_SESSION_CA)),
)
class SingleSessionAllenCaTime(SingleSessionAllenCa):
    """A single mouse 30Hz calcium events dataset during the allen MOVIE1 stimulus.

    A dataset of a single mouse 30Hz calcium events from the excitatory neurons in the primary visual cortex
    during the 10 repeats of the MOVIE1 stimulus in session type A. The preprocessed data from *Deitch et al. (2021) are used.
    There is no behavioral variable used as the continuous label.

    Args:
        session_id: The integer value to pick a session among 4 sessions with the largest number of recorded neruons. Choose between 0-3.
        frame_feature_path: The path of the movie frame features.

    """

    @property
    def continuous_index(self):
        return torch.arange(len(self))


@parametrize(
    "allen-movie1-ca-single-session-decoding-{session_id}-repeat-{repeat_no}-{split_flag}",
    session_id=range(len(_SINGLE_SESSION_CA)),
    repeat_no=np.arange(10),
    split_flag=["train", "test"],
)
class SingleSessionAllenCaDecoding(cebra.data.SingleSessionDataset):
    """A single mouse 30Hz calcium events dataset during the allen MOVIE1 stimulus with train/test splits.

    A dataset of a single mouse 30Hz calcium events from the excitatory neurons in the primary visual cortex
    during the 10 repeats of the MOVIE1 stimulus in session type A. The preprocessed data from *Deitch et al. (2021) are used.
    The continuous labels corresponding to a DINO embedding of each stimulus frame.
    A neural recording during the chosen repeat is used as a test set and the remaining 9 repeats are used as a train set.

    Args:
        session_id: The integer value to pick a session among 4 sessions with the largest number of recorded neruons. Choose between 0-3.
        repeat_no: The nth repeat to use as the test set. Choose between 0-9.
        split_flag: The `train`/`test` split to load.
        frame_feature_path: The path of the movie frame features.
        pca: If true, 32 principal components of the PCA transformed calcium data are used as neural input. Default value is `False`.

    """

    def __init__(
        self,
        session_id: int,
        repeat_no: int,
        split_flag: str,
        frame_feature_path: str = pathlib.Path(_DEFAULT_DATADIR) / "allen" /
        "features" / "allen_movies" / "vit_base" / "8" /
        "movie_one_image_stack.npz" / "testfeat.pth",
        pca: bool = False,
    ):
        self.path = _SINGLE_SESSION_CA[session_id]
        traces = scipy.io.loadmat(self.path)
        if pca:
            pca_ = PCA()
            neural = pca_.fit_transform(
                traces["filtered_traces_days_events"][0,
                                                      0].transpose(1,
                                                                   0))[:, :32]
        else:
            neural = traces["filtered_traces_days_events"][0, 0].transpose(1, 0)

        test_idx = np.arange(900 * repeat_no, 900 * (repeat_no + 1))
        train_idx = np.delete(np.arange(9000), test_idx)
        frame_feature = torch.load(frame_feature_path)
        if split_flag == "train":
            self.neural = torch.from_numpy(neural[train_idx]).float()
            self.index = frame_feature.repeat(9, 1)
        elif split_flag == "test":
            self.neural = torch.from_numpy(neural[test_idx]).float()
            self.index = frame_feature

    def __len__(self):
        return self.neural.size(0)

    @property
    def continuous_index(self):
        return self.index

    @property
    def input_dimension(self):
        return self.neural.size(1)

    def __getitem__(self, index):
        index = self.expand_index(index)
        return self.neural[index].transpose(2, 1)


@parametrize(
    "allen-movie1-ca-single-session-leave2out-{session_id}-repeat-{repeat_no}-{split_flag}",
    session_id=range(len(_SINGLE_SESSION_CA)),
    repeat_no=[0, 2, 4, 6, 8],
    split_flag=["train", "valid", "test"],
)
class SingleSessionAllenCaDecodingLeave2Out(cebra.data.SingleSessionDataset):

    def __init__(
        self,
        session_id,
        repeat_no,
        split_flag,
        frame_feature_path=pathlib.Path(_DEFAULT_DATADIR) / "allen" /
        "features" / "allen_movies" / "vit_base" / "8" /
        "movie_one_image_stack.npz" / "testfeat.pth",
    ):
        self.path = _SINGLE_SESSION_CA[session_id]
        traces = scipy.io.loadmat(self.path)
        neural = traces["filtered_traces_days_events"][0, 0].transpose(1, 0)

        valid_idx = np.arange(900 * repeat_no, 900 * (repeat_no + 1))
        test_idx = np.arange(900 * (repeat_no + 1), 900 * (repeat_no + 2))
        train_idx = np.delete(np.arange(9000),
                              np.concatenate([valid_idx, test_idx]))
        frame_feature = torch.load(frame_feature_path)
        if split_flag == "train":
            self.neural = torch.from_numpy(neural[train_idx]).float()
            self.index = frame_feature.repeat(9, 1)
        elif split_flag == "valid":
            self.neural = torch.from_numpy(neural[valid_idx]).float()
            self.index = frame_feature
        elif split_flag == "test":
            self.neural = torch.from_numpy(neural[test_idx]).float()
            self.index = frame_feature
        else:
            raise ValueError(split_flag)

    def __len__(self):
        return self.neural.size(0)

    @property
    def continuous_index(self):
        return self.index

    @property
    def input_dimension(self):
        return self.neural.size(1)

    def __getitem__(self, index):
        index = self.expand_index(index)
        return self.neural[index].transpose(2, 1)


@parametrize(
    "allen-movie1-ca-multi-session-decoding-repeat-{repeat_no}-{split_flag}",
    repeat_no=np.arange(10),
    split_flag=["train", "test"],
)
class MultiSessionAllenCaDecoding(cebra.data.DatasetCollection):

    def __init__(self, repeat_no, split_flag):
        super().__init__(
            *[
                init(
                    f"allen-movie1-ca-single-session-decoding-{session_id}-repeat-{repeat_no}-{split_flag}"
                ) for session_id in range(4)
            ],)


@parametrize(
    "allen-movie1-ca-multi-session-leave2out-repeat-{repeat_no}-{split_flag}",
    repeat_no=[0, 2, 4, 6, 8],
    split_flag=["train", "valid", "test"],
)
class MultiSessionAllenCaLeave2Out(cebra.data.DatasetCollection):

    def __init__(self, repeat_no, split_flag):
        super().__init__(
            *[
                init(
                    f"allen-movie1-ca-single-session-leave2out-{session_id}-repeat-{repeat_no}-{split_flag}"
                ) for session_id in range(4)
            ],)


@parametrize(
    "allen-movie1-ca-single-session-decoding-corrupt-{session_id}-repeat-{repeat_no}-{split_flag}",
    session_id=range(len(_SINGLE_SESSION_CA)),
    repeat_no=[9],
    split_flag=["train", "test"],
)
class SingleSessionAllenCaDecodingCorrupted(cebra.data.SingleSessionDataset):
    """A corrupted single mouse 30Hz calcium events dataset during the allen MOVIE1 stimulus with train/test splits.

    A dataset of a single mouse 30Hz calcium events from the excitatory neurons
    in the primary visual cortex during the 10 repeats of the MOVIE1 stimulus
    in session type A. The preprocessed data from *Deitch et al. (2021) are used.
    The continuous labels corresponding to a DINO embedding of each stimulus frame,
    but in randomly shuffled order.
    A neural recording during the chosen repeat is used as a test set and the
    remaining 9 repeats are used as a train set.

    Args:
        session_id: The integer value to pick a session among 4 sessions with the
            largest number of recorded neruons. Choose between 0-3.
        repeat_no: The nth repeat to use as the test set. Choose between 0-9.
        split_flag: The `train`/`test` split to load.
        frame_feature_path: The path of the movie frame features.

    """

    def __init__(
        self,
        session_id: int,
        repeat_no: int,
        split_flag: str,
        frame_feature_path: str = pathlib.Path(_DEFAULT_DATADIR) / "allen" /
        "features" / "allen_movies" / "vit_base" / "8" /
        "movie_one_image_stack.npz" / "testfeat.pth",
        pca: bool = False,
    ):
        self.path = _SINGLE_SESSION_CA[session_id]
        traces = scipy.io.loadmat(self.path)
        if pca:
            pca_ = PCA()
            neural = pca_.fit_transform(
                traces["filtered_traces_days_events"][0,
                                                      0].transpose(1,
                                                                   0))[:, :32]
        else:
            neural = traces["filtered_traces_days_events"][0, 0].transpose(1, 0)

        test_idx = np.arange(900 * repeat_no, 900 * (repeat_no + 1))
        train_idx = np.delete(np.arange(9000), test_idx)
        frame_feature = torch.load(frame_feature_path)
        rng = np.random.Generator(np.random.PCG64(111))
        if split_flag == "train":
            self.neural = torch.from_numpy(neural[train_idx]).float()
            self.frame_index = np.arange(900)
            rng.shuffle(self.frame_index)
            self.index = frame_feature.repeat(9,
                                              1)[np.tile(self.frame_index, 9)]
        elif split_flag == "test":
            self.neural = torch.from_numpy(neural[test_idx]).float()
            self.index = frame_feature
            self.frame_index = np.arange(900)
            rng.shuffle(self.frame_index)
            self.index = frame_feature[self.frame_index]

    def __len__(self):
        return self.neural.size(0)

    @property
    def continuous_index(self):
        return self.index

    @property
    def input_dimension(self):
        return self.neural.size(1)

    def __getitem__(self, index):
        index = self.expand_index(index)
        return self.neural[index].transpose(2, 1)



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Matplotlib interface to CEBRA."""
import abc
from collections.abc import Iterable
from typing import List, Literal, Optional, Tuple, Union

import matplotlib.axes
import matplotlib.cm
import matplotlib.colors
import matplotlib.figure
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import sklearn.utils.validation
import torch

from cebra import CEBRA
from cebra.helper import requires_package_version


def _register_colormap():
    """Register colormaps used in CEBRA plotting.

    Currently registered:
        * Logo colors (``cebra``)
    """

    if "cebra" not in matplotlib.colormaps:
        matplotlib.colormaps.register(
            matplotlib.colors.LinearSegmentedColormap.from_list(
                "", ["#4854e9", "#6235e0", "#a045e8", "#bf1bb9", "#d4164f"]),
            name="cebra",
            force=False,
        )


class _BasePlot:
    """Base plotting class.

    Attributes:
        axis: Optional axis to create the plot on.
        figsize: Figure width and height in inches.
        dpi: Figure resolution.
    """

    def __init__(self, axis: Optional[matplotlib.axes.Axes], figsize: tuple,
                 dpi: int):
        if axis is None:
            self.fig = plt.figure(figsize=figsize, dpi=dpi)

    @abc.abstractmethod
    def _define_ax(
            self, axis: Optional[matplotlib.axes.Axes]) -> matplotlib.axes.Axes:
        raise NotImplementedError()

    @abc.abstractmethod
    def plot(self, **kwargs) -> matplotlib.axes.Axes:
        raise NotImplementedError()


class _TemperaturePlot(_BasePlot):
    """Plot temperature evolution during model training.

    Attributes:
        model: The (trained) CEBRA model.
        color: Line color.
        linewidth: Line width.
        x_label: A boolean that specifies if the x-axis label should be displayed.
        y_label: A boolean that specifies if the y-axis label should be displayed.
        axis: Optional axis to create the plot on.
        figsize: Figure width and height in inches.
        dpi: Figure resolution.
    """

    def __init__(
        self,
        model: CEBRA,
        color: str,
        linewidth: float,
        x_label: bool,
        y_label: bool,
        axis: Optional[matplotlib.axes.Axes],
        figsize: tuple,
        dpi: int,
    ):
        super().__init__(axis, figsize, dpi)
        self.ax = self._define_ax(axis)
        sklearn.utils.validation.check_is_fitted(model, "n_features_")
        self.model = model
        self.color = color
        self.linewidth = linewidth
        self.x_label = x_label
        self.y_label = y_label

    def _define_ax(
            self, axis: Optional[matplotlib.axes.Axes]) -> matplotlib.axes.Axes:
        """Define the ax on which to generate the plot.

        Args:
            axis: A required ``matplotlib.axes.Axes``. If None, then add an axis to the current figure.

        Returns:
            A ``matplotlib.axes.Axes`` on which to generate the plot.
        """
        if axis is None:
            self.ax = self.fig.add_subplot()
        else:
            self.ax = axis
        return self.ax

    def plot(self, **kwargs) -> matplotlib.axes.Axes:
        """Plot the temperature.

        Returns:
            The axis :py:meth:`matplotlib.axes.Axes.axis` of the plot.
        """
        self.ax.plot(
            self.model.state_dict_["log"]["temperature"],
            c=self.color,
            linewidth=self.linewidth,
            **kwargs,
        )
        if self.y_label:
            self.ax.set_ylabel("Temperature")
        if self.x_label:
            self.ax.set_xlabel("Steps")

        return self.ax


class _LossPlot(_BasePlot):
    """Plot loss evolution during model training.

    Attributes:
        model: The (trained) CEBRA model.
        label: The legend for the trace.
        color: Line color.
        linewidth: Line width.
        x_label: A boolean that specifies if the x-axis label should be displayed.
        y_label: A boolean that specifies if the y-axis label should be displayed.
        axis: Optional axis to create the plot on.
        figsize: Figure width and height in inches.
        dpi: Figure resolution.
        kwargs: Optional arguments to customize the plots.
    """

    def __init__(
        self,
        model: CEBRA,
        label: Optional[Union[str, int, float]],
        color: str,
        linewidth: float,
        x_label: bool,
        y_label: bool,
        axis: Optional[matplotlib.axes.Axes],
        figsize: tuple,
        dpi: int,
    ):
        super().__init__(axis, figsize, dpi)
        self.ax = self._define_ax(axis)
        sklearn.utils.validation.check_is_fitted(model, "n_features_")
        self.model = model
        self.label = label
        self.color = color
        self.linewidth = linewidth
        self.x_label = x_label
        self.y_label = y_label

    def _define_ax(
            self, axis: Optional[matplotlib.axes.Axes]) -> matplotlib.axes.Axes:
        """Define the ax on which to generate the plot.

        Args:
            axis: A required ``matplotlib.axes.Axes``. If None, then add an axis to the current figure.

        Returns:
            A ``matplotlib.axes.Axes`` on which to generate the plot.
        """
        if axis is None:
            self.ax = self.fig.add_subplot()
        else:
            self.ax = axis
        return self.ax

    def plot(self, **kwargs) -> matplotlib.axes.Axes:
        """Plot the loss.

        Returns:
            The axis :py:meth:`matplotlib.axes.Axes.axis` of the plot.
        """
        self.ax.plot(
            self.model.state_dict_["loss"],
            c=self.color,
            linewidth=self.linewidth,
            label=self.label,
            **kwargs,
        )
        if self.y_label:
            loss = self.model.__dict__["criterion"]
            if loss == "infonce":
                loss = "InfoNCE"
            self.ax.set_ylabel(f"{loss} Loss")
        if self.x_label:
            self.ax.set_xlabel("Steps")

        if self.label is not None:
            self.ax.legend()

        return self.ax


class _EmbeddingPlot(_BasePlot):
    """Plot a CEBRA embedding in a 3D or 2D dimensional space.

    Attributes:
        embedding: A matrix containing the feature representation computed with CEBRA.
        embedding_labels: The labels used to map the data to color. It can be a vector that is the
            same sample size as the embedding, associating a value to each of the sample, either discrete
            or continuous or string, either `time`, then the labels while color the embedding based on
            temporality, or a string that can be interpreted as a RGB(A) color, then the embedding will be
            uniformly display with that unique color.
        ax: Optional axis to create the plot on.
        idx_order: A tuple (x, y, z) or (x, y) that maps a dimension in the data to a dimension in the 3D/2D
            embedding. The simplest form is (0, 1, 2) or (0, 1) but one might want to plot either those
            dimensions differently (e.g., (1, 0, 2)) or other dimensions from the feature representation
            (e.g., (2, 4, 5)).
        markersize: The marker size.
        alpha: The marker blending, between 0 (transparent) and 1 (opaque).
        cmap: The Colormap instance or registered colormap name used to map scalar data to colors.
            It will be ignored if `embedding_labels` is set to a valid RGB(A).
        title: The title on top of the embedding.
        axis: Optional axis to create the plot on.
        figsize: Figure width and height in inches.
        dpi: Figure resolution.

    """

    def __init__(
        self,
        embedding: Union[npt.NDArray, torch.Tensor],
        embedding_labels: Optional[Union[npt.NDArray, torch.Tensor, str]],
        idx_order: Optional[Tuple[int]],
        markersize: float,
        alpha: float,
        cmap: str,
        title: str,
        axis: Optional[matplotlib.axes.Axes],
        figsize: tuple,
        dpi: int,
    ):
        super().__init__(axis, figsize, dpi)
        self._define_plot_dim(embedding, idx_order)
        self._define_ax(axis)
        self.embedding = embedding
        self.embedding_labels = embedding_labels
        self.idx_order = self._define_idx_order(idx_order)
        self.markersize = markersize
        self.alpha = alpha
        self.cmap = cmap
        self.title = title

    def _define_plot_dim(
        self,
        embedding: Union[npt.NDArray, torch.Tensor],
        idx_order: Optional[Tuple[int]],
    ):
        """Define the dimension of the embedding plot, either 2D or 3D, by setting ``_is_plot_3d``.

        If the embedding dimension is equal or higher to 3:

            * If ``idx_order`` is not provided, the plot will be 3D by default.
            * If ``idx_order`` is provided, if it has 3 dimensions, the plot will be 3D, if only 2 dimensions
            are provided, the plot will be 2D.

        If the embedding dimension is equal to 2:

            * If ``idx_order`` is not provided, the plot will be 2D by default.
            * If ``idx_order`` is provided, if it has 3 dimensions, the plot will be 3D, if 2 dimensions
            are provided, the plot will be 2D.

        This is supposing that the dimensions provided to ``idx_order`` are in the range of the number of
        dimensions of the embedding (i.e., between 0 and :py:attr:`cebra.CEBRA.output_dimension` -1).

        Args:
            embedding: A matrix containing the feature representation computed with CEBRA.
            idx_order: A tuple (x, y, z) or (x, y) that maps a dimension in the data to a dimension in the 3D/2D
                embedding. The simplest form is (0, 1, 2) or (0, 1) but one might want to plot either those
                dimensions differently (e.g., (1, 0, 2)) or other dimensions from the feature representation
                (e.g., (2, 4, 5)).
        """
        if (idx_order is None and
                embedding.shape[1] == 2) or (idx_order is not None and
                                             len(idx_order) == 2):
            self._is_plot_3d = False
        elif (idx_order is None and
              embedding.shape[1] >= 3) or (idx_order is not None and
                                           len(idx_order) == 3):
            self._is_plot_3d = True
        else:
            raise ValueError(
                f"Invalid embedding dimension, expects 2D or more, got {self.embedding.shape[1]}"
            )

    def _define_ax(self, axis: Optional[matplotlib.axes.Axes]):
        """Define the ax on which to generate the plot.

        Args:
            axis: A required ``matplotlib.axes.Axes``. If None, then add an axis to the current figure.

        Returns:
            A ``matplotlib.axes.Axes`` on which to generate the plot.
        """
        if axis is None:
            if self._is_plot_3d:
                self.ax = self.fig.add_subplot(projection="3d")
            else:
                self.ax = self.fig.add_subplot()
        else:
            self.ax = axis

    def _define_idx_order(self, idx_order: Optional[Tuple[int]]) -> Tuple[int]:
        """Check that the index order has a valid number of dimensions compared to the number of
        dimensions of the embedding.

        Args:
            idx_order: A tuple (x, y, z) or (x, y) that maps a dimension in the data to a dimension in the 3D/2D
                embedding. The simplest form is (0, 1, 2) or (0, 1) but one might want to plot either those
                dimensions differently (e.g., (1, 0, 2)) or other dimensions from the feature representation
                (e.g., (2, 4, 5)).

        Returns:
            The index order for the corresponding 2D or 3D plot.
        """

        if idx_order is None:
            if self._is_plot_3d:
                idx_order = (0, 1, 2)
            else:
                idx_order = (0, 1)
        else:
            # If the idx_order was provided by the user
            self._check_valid_dimensions(idx_order)
        return idx_order

    def _check_valid_dimensions(self, idx_order: Tuple[int]):
        """Check that provided dimensions are valid.

        The provided dimensions need to be 2 if the plot is set to a 2D plot and 3 if it is set to 3D.
        The dimensions values should be in the range of the embedding dimensionality.

        Args:
            idx_order: A tuple (x, y, z) or (x, y) that maps a dimension in the data to a dimension in the 3D/2D
                embedding. The simplest form is (0, 1, 2) or (0, 1) but one might want to plot either those
                dimensions differently (e.g., (1, 0, 2)) or other dimensions from the feature representation
                (e.g., (2, 4, 5)).
        """
        # Check size validity
        if self._is_plot_3d and len(idx_order) != 3:
            raise ValueError(
                f"idx_order must contain 3 dimension values, got {len(idx_order)}."
            )
        elif not self._is_plot_3d and len(idx_order) != 2:
            raise ValueError(
                f"idx_order must contain 2 dimension values, got {len(idx_order)}."
            )

        # Check value validity
        for dim in idx_order:
            if dim < 0 or dim > self.embedding.shape[1] - 1:
                raise ValueError(
                    f"List of dimensions to plot is invalid, got {idx_order}, with {dim} invalid."
                    f"Values should be between 0 and {self.embedding.shape[1]}."
                )

    def _plot_3d(self,
                 grey_fig: bool = False,
                 **kwargs) -> matplotlib.axes.Axes:
        """Plot the embedding in 3d.

        Args:
            grey_fig: Set the title and edge color to grey, to be visible on both white and black
                backgrounds.


        Returns:
            The axis :py:meth:`matplotlib.axes.Axes.axis` of the plot.
        """

        idx1, idx2, idx3 = self.idx_order
        self.ax.scatter(
            xs=self.embedding[:, idx1],
            ys=self.embedding[:, idx2],
            zs=self.embedding[:, idx3],
            c=self.embedding_labels,
            cmap=self.cmap,
            alpha=self.alpha,
            s=self.markersize,
            **kwargs,
        )

        self.ax.grid(False)
        self.ax.xaxis.pane.fill = False
        self.ax.yaxis.pane.fill = False
        self.ax.zaxis.pane.fill = False
        self.ax.xaxis.pane.set_edgecolor("w")
        self.ax.yaxis.pane.set_edgecolor("w")
        self.ax.zaxis.pane.set_edgecolor("w")
        self.ax.set_title(self.title, y=1.0, pad=-10)

        if grey_fig:
            self.ax.xaxis.pane.set_edgecolor("grey")
            self.ax.yaxis.pane.set_edgecolor("grey")
            self.ax.zaxis.pane.set_edgecolor("grey")

        return self.ax

    def _plot_2d(self, **kwargs) -> matplotlib.axes.Axes:
        """Plot the embedding in 2d.

        Returns:
            The axis :py:meth:`matplotlib.axes.Axes.axis` of the plot.
        """

        idx1, idx2 = self.idx_order
        self.ax.scatter(
            x=self.embedding[:, idx1],
            y=self.embedding[:, idx2],
            c=self.embedding_labels,
            cmap=self.cmap,
            alpha=self.alpha,
            s=self.markersize,
            **kwargs,
        )

        return self.ax

    def plot(self, **kwargs) -> matplotlib.axes.Axes:
        """Plot the embedding.

        Note:
            To set the entire figure to grey, you should add that snippet of code:

            >>> from matplotlib import rcParams
            >>> rcParams['xtick.color'] = 'grey'
            >>> rcParams['ytick.color'] = 'grey'
            >>> rcParams['axes.labelcolor'] = 'grey'
            >>> rcParams['axes.edgecolor'] = 'grey'
            >>> rcParams['axes.titlecolor'] = 'grey'

        Returns:
            The axis :py:meth:`matplotlib.axes.Axes.axis` of the plot.
        """
        if isinstance(self.embedding_labels, str):
            if self.embedding_labels == "time":
                self.embedding_labels = np.arange(self.embedding.shape[0])
            elif not matplotlib.colors.is_color_like(self.embedding_labels):
                raise ValueError(
                    f"Embedding labels invalid: provide a list of index or a valid str (time or valid colorname), got {self.embedding_labels}."
                )
            self.cmap = None
        elif isinstance(self.embedding_labels, Iterable):
            if len(self.embedding_labels) != self.embedding.shape[0]:
                raise ValueError(
                    f"Invalid embedding labels: the labels vector should have the same number "
                    f"of samples as the embedding, got {len(self.embedding_labels)}, "
                    f"expected {self.embedding.shape[0]}.")
            if self.embedding_labels.ndim > 1:
                raise NotImplementedError(
                    f"Invalid embedding labels: plotting does not support multiple sets of labels, got {self.embedding_labels.ndim}."
                )

        if self._is_plot_3d:
            self.ax = self._plot_3d(**kwargs)
        else:
            self.ax = self._plot_2d(**kwargs)
        if isinstance(self.ax, matplotlib.axes._axes.Axes):
            self.ax.set_title(self.title)

        return self.ax


class _ConsistencyPlot(_BasePlot):
    """Plot a consistency matrix from the scores obtained with :py:func:`~.consistency_score`.

    Attributes:
        scores: List of consistency scores obtained by comparing a set of CEBRA embeddings using :py:func:`~.consistency_score`.
        pairs: Optional list of the pairs of datasets/runs whose embeddings were compared in ``scores``.
        datasets: Optional list of the datasets whose embeddings were compared in ``scores``. Each dataset is present once only in the list.
        ax: Optional axis to create the plot on.
        cmap: Color map to use to color the m
        text_color: If None, then the ``scores`` values are not displayed, else, it corresponds to the color
            of the values displayed inside the grid.
        colorbar_label: If None, then the color bar is not shown, else, it defines the corresponding title.
        title: The title on top of the confusion matrix.
        figsize: Figure width and height in inches.
        dpi: Figure resolution.

    """

    def __init__(
        self,
        scores: Union[torch.Tensor, list, npt.NDArray],
        datasets: Optional[Union[list, npt.NDArray]],
        pairs: Optional[Union[list, npt.NDArray]],
        cmap: str,
        text_color: Optional[str],
        colorbar_label: Optional[str],
        title: Optional[str],
        axis: Optional[matplotlib.axes.Axes],
        figsize: tuple,
        dpi: int,
    ):
        super().__init__(axis, figsize, dpi)
        self._define_ax(axis)
        scores = self._check_array(scores)
        # Check the values dimensions
        if scores.ndim >= 2:
            raise ValueError(
                f"Invalid scores dimensions, expect 1D, got {scores.ndim}D.")

        self.labels = self._compute_labels(scores,
                                           pairs=pairs,
                                           datasets=datasets)
        self.scores = self._to_heatmap_format(scores)
        self.cmap = cmap
        self.text_color = text_color
        self.colorbar_label = colorbar_label
        self.title = title

    def _define_ax(
            self, axis: Optional[matplotlib.axes.Axes]) -> matplotlib.axes.Axes:
        """Define the ax on which to generate the matrix.

        Args:
            axis: A required ``matplotlib.axes.Axes``. If None, then add an axis to the current figure.

        Returns:
            A ``matplotlib.axes.Axes`` on which to generate the plot.
        """
        if axis is None:
            self.ax = self.fig.add_subplot()
        else:
            self.ax = axis
        return self.ax

    def _check_array(
            self, array: Union[torch.Tensor, list, npt.NDArray]) -> npt.NDArray:
        """Check the ``array`` type and either convert to :py:func:`numpy.array` or raise an
        error.

        Args:
            array: Data structure to convert to :py:func:`numpy.array`.

        Returns:
            A :py:func:`numpy.array`.
        """
        if not isinstance(array, np.ndarray):
            if isinstance(array, list):
                array = np.array(array)
            elif isinstance(array, torch.Tensor):
                array = array.numpy()
            else:
                raise ValueError(
                    f"Invalid values format, expect a numpy.array, torch.Tensor or list, got {type(array)}."
                )
        return array

    def _compute_labels(
        self,
        scores: Union[torch.Tensor, list, npt.NDArray],
        pairs: Union[list, npt.NDArray],
        datasets: Union[list, npt.NDArray],
    ) -> npt.NDArray:
        """Define the x- and y-labels to use for the confusion matrix.

        The labels qualify the embeddings that were compared and are selected based on ``pairs`` and ``datasets``.
        Note that as comparison is one-to-one, the confusion matrix is a square and
        the labels are similar on the x- and y-axis. First we check if the number of datasets compared to
        the number of scores is consistent to a between-datasets comparison. As we do not compute
        self-consistency, we expect :math:`#scores = #datasets^2 - #datasets` as well as finding the datasets
        IDs present in ``datasets`` also in ``pairs``. If that is not the case, we check if the number of
        pairs in ``pairs`` is consistent with the number of scores in ``scores``. For that, we expect
        :math:`#scores = #pairs` and the number of unique dataset IDs in ``pairs`` is such that, similarly,
        :math:`#scores = #set(pairs)^2 - #set(pairs)`, with :math:`set(pairs)` the set of unique dataset IDs
        in ``pairs``.

        Args:
            scores: List of consistency scores obtained by comparing a set of CEBRA embeddings using
                :py:func:`~.consistency_score`.
            pairs: Optional list of the pairs of datasets/runs whose embeddings were compared in ``scores``.
            datasets: Optional list of the datasets whose embeddings were compared in ``scores``. Each dataset
                is present once only in the list.

        Returns:
            The labels to use.
        """
        if datasets is None or pairs is None:
            raise ValueError(
                "Missing datasets or pairs, provide both of them to plot consistency, "
                "got either both or one of them  set to None.")
        else:
            datasets = self._check_array(datasets)
            pairs = self.pairs = self._check_array(pairs)

            if len(pairs.shape) == 2:
                compared_items = list(sorted(set(pairs[:, 0])))
            elif len(pairs.shape) == 3:
                compared_items = list(sorted(set(pairs[0, :, 0])))

            # between-datasets comparison
            if len(scores) == len(datasets)**2 - len(datasets) and all(
                    mouse in compared_items for mouse in datasets):
                self.labels = datasets
            # between-runs comparison
            elif len(scores) == len(compared_items)**2 - len(
                    compared_items) and not any(mouse in compared_items
                                                for mouse in datasets):
                self.labels = np.array(compared_items)
            else:
                raise ValueError(
                    f"Shape of the scores, datasets and pairs do not match, got scores:{scores.shape}, datasets:{len(datasets)} and pairs:{len(pairs)}."
                )

        return self.labels

    def _to_heatmap_format(
            self, values: Union[torch.Tensor, list,
                                npt.NDArray]) -> npt.NDArray:
        """Transform ``values`` to a compatible format for :py:func:`matplotlib.pyplot.imshow`.

        Args:
            values: An array of values to transform to a compatible format, meaning, square format,
                of size `len(self.labels)`, with NaNs on the diagonal. We also transform the values
                to percentages.

        Returns:
            A :py:func:`numpy.array` of shape ``(len(self.labels), len(self.labels))``.

        """
        if values.ndim == 1:
            values = np.expand_dims(values, axis=0)

        values = np.concatenate(values)

        pairs = self.pairs

        if pairs.ndim == 3:
            pairs = pairs[0]

        assert len(pairs) == len(values), (self.pairs.shape, len(values))
        score_dict = {tuple(pair): value for pair, value in zip(pairs, values)}

        # NOTE(stes): Never used, might be possible to remove.
        #if self.labels is None:
        #    n_grid = self.score

        heatmap_values = np.zeros((len(self.labels), len(self.labels)))

        heatmap_values[:] = float("nan")
        for i, label_i in enumerate(self.labels):
            for j, label_j in enumerate(self.labels):
                if i == j:
                    heatmap_values[i, j] = float("nan")
                else:
                    heatmap_values[i, j] = score_dict[label_i, label_j]

        return np.minimum(heatmap_values * 100, 99)

    def _create_text(self):
        """Create the text to add in the confusion matrix grid and the title."""
        if self.text_color is not None:
            for (i, j), z in np.ndenumerate(self.scores):
                if z == z:  # not NaN
                    self.ax.text(
                        j,
                        i,
                        "{:0.1f}".format(z),
                        ha="center",
                        va="center",
                        color=self.text_color,
                    )

        if self.title is not None:
            self.ax.set_title(self.title)

    def _create_colorbar(self, im):
        """Create color bar based on the provided ``colorbar_label``."""
        if self.colorbar_label is not None:
            cbar = self.ax.figure.colorbar(im, ax=self.ax)
            cbar.outline.set_visible(False)
            cbar.ax.set_ylabel(self.colorbar_label, rotation=-90, va="bottom")

    def _create_labels(self):
        """Create and add the labels to the confusion matrix."""
        # Show labels
        if self.labels is not None:
            self.ax.set_xticks(np.arange(self.scores.shape[1]),
                               labels=self.labels)
            self.ax.set_yticks(np.arange(self.scores.shape[0]),
                               labels=self.labels)

        # Remove labels ticks and pivot y-labels
        self.ax.tick_params(left=False,
                            bottom=False,
                            labelleft=True,
                            labelbottom=True)
        plt.setp(self.ax.get_yticklabels(),
                 rotation=90,
                 ha="center",
                 rotation_mode="anchor")
        self.ax.tick_params(axis="y", which="major", pad=10)

    def plot(self, **kwargs) -> matplotlib.axes.Axes:
        """Plot the consistency matrix.

        Returns:
            The axis :py:meth:`matplotlib.axes.Axes.axis` of the matrix.
        """
        im = self.ax.imshow(
            self.scores,
            cmap=self.cmap,
            **kwargs,
        )

        self._create_text()
        self._create_colorbar(im)
        self._create_labels()

        # Turn spines off and create white grid.
        self.ax.spines[:].set_visible(False)
        self.ax.tick_params(which="minor", bottom=False, left=False)

        return self.ax


def plot_overview(
    model: CEBRA,
    X: Union[npt.NDArray, torch.Tensor],
    loss_kwargs: dict = {},
    temperature_kwargs: dict = {},
    embedding_kwargs: dict = {},
    figsize: tuple = (15, 4),
    dpi: int = 100,
    **kwargs,
) -> Tuple[matplotlib.figure.Figure, Tuple[
        matplotlib.axes.Axes, matplotlib.axes.Axes, matplotlib.axes.Axes]]:
    """Plot an overview of a trained CEBRA model.

    Args:
        model: The (fitted) CEBRA model to analyze to compute the different plots.
        X: The data matrix on which to compute the embedding to plot.
        figsize: Figure width and height in inches.
        dpi: Figure resolution.
        kwargs: Optional arguments to customize the plots.

    Returns:
        The axis :py:meth:`matplotlib.axes.Axes.axis` of the plot.

    Example:

        >>> import cebra
        >>> import numpy as np
        >>> X = np.random.uniform(0, 1, (100, 50))
        >>> y = np.random.uniform(0, 10, (100, 5))
        >>> cebra_model = cebra.CEBRA(max_iterations=10)
        >>> cebra_model.fit(X, y)
        CEBRA(max_iterations=10)
        >>> fig = cebra.plot_overview(cebra_model, X, embedding_kwargs={"embedding_labels":"time"})

    """

    fig = plt.figure(figsize=figsize, dpi=dpi)
    grid = plt.GridSpec(4, 2, wspace=0, hspace=0.2, figure=fig)

    # Plot the loss
    ax1 = fig.add_subplot(grid[:2, 0])
    ax1 = plot(
        model,
        which="loss",
        ax=ax1,
        x_label=False,
        linewidth=1,
        figure=fig,
        **loss_kwargs,
        **kwargs,
    )
    ax2 = fig.add_subplot(grid[2:, 0], sharex=ax1)

    # Plot the temperature
    ax2 = plot(model,
               which="temperature",
               ax=ax2,
               linewidth=1,
               **temperature_kwargs,
               **kwargs)
    plt.setp(ax1.get_xticklabels(), visible=False)

    # Compute and plot the embedding of X
    ax3 = fig.add_subplot(grid[:, 1:], projection="3d")
    ax3 = plot(
        model,
        which="embedding",
        X=X,
        ax=ax3,
        markersize=0.01,
        **embedding_kwargs,
        **kwargs,
    )

    return fig, (ax1, ax2, ax3)


def plot(
    model: CEBRA,
    which: str = Literal["loss", "temperature", "embedding"],
    X: Optional[Union[npt.NDArray, torch.Tensor]] = None,
    ax: Optional[matplotlib.axes.Axes] = None,
    **kwargs,
) -> matplotlib.axes.Axes:
    """Plot the required information.

    Args:
        model: The (fitted) CEBRA model to analyze to compute the different plots.
        which: The required information to plot.
        X: The data matrix on which to compute the embedding to plot, needed only to display an embedding.
        ax: Optional axis to create the plot on.
        kwargs: Optional arguments to customize the plots or provide the embedding labels.

    Example:

        >>> import cebra
        >>> import numpy as np
        >>> X = np.random.uniform(0, 1, (100, 50))
        >>> y = np.random.uniform(0, 10, (100, 5))
        >>> cebra_model = cebra.CEBRA(max_iterations=10)
        >>> cebra_model.fit(X, y)
        CEBRA(max_iterations=10)
        >>> ax = cebra.plot(cebra_model, which="temperature")

    Returns:
        The axis :py:meth:`matplotlib.axes.Axes.axis` of the plot.
    """
    if which == "loss":
        ax = plot_loss(model, ax=ax, **kwargs)
    elif which == "temperature":
        ax = plot_temperature(model, ax=ax, **kwargs)
    elif which == "embedding":
        if X is None:
            raise ValueError(
                "No data: a data matrix to compute the embedding on needs to be provided."
            )
        embedding = model.transform(X)

        ax = plot_embedding(embedding=embedding, ax=ax, **kwargs)
    else:
        raise ValueError(
            f"Invalid value for `which`: got {which}, expect loss, temperature or embedding."
        )

    return ax


def plot_temperature(
    model: CEBRA,
    ax: Optional[matplotlib.axes.Axes] = None,
    color: str = "dodgerblue",
    linewidth: int = 1,
    x_label: bool = True,
    y_label: bool = True,
    figsize: tuple = (7, 4),
    dpi: float = 100,
    **kwargs,
) -> matplotlib.axes.Axes:
    """Plot the evolution of the :py:attr:`~.CEBRA.temperature` hyperparameter during model training.

    Note:
        It will vary only when using :py:attr:`~.CEBRA.temperature_mode`, else the :py:attr:`~.CEBRA.temperature` stays constant.

    The function makes use of :py:func:`matplotlib.pyplot.plot` and parameters from that function can be provided
    as part of ``kwargs``.

    Args:
        model: The (trained) CEBRA model.
        ax: Optional axis to create the plot on.
        color: Line color.
        linewidth: Line width.
        x_label: A boolean that specifies if the x-axis label should be displayed.
        y_label: A boolean that specifies if the y-axis label should be displayed.
        figsize: Figure width and height in inches.
        dpi: Figure resolution.
        kwargs: Optional arguments to customize the plots. See :py:func:`matplotlib.pyplot.plot` documentation for more
            details on which arguments to use.

    Example:

        >>> import cebra
        >>> import numpy as np
        >>> X = np.random.uniform(0, 1, (100, 50))
        >>> y = np.random.uniform(0, 10, (100, 5))
        >>> cebra_model = cebra.CEBRA(max_iterations=10)
        >>> cebra_model.fit(X, y)
        CEBRA(max_iterations=10)
        >>> ax = cebra.plot_temperature(cebra_model)

    Returns:
        The axis :py:meth:`matplotlib.axes.Axes.axis` of the plot.

    """

    return _TemperaturePlot(
        model=model,
        axis=ax,
        color=color,
        linewidth=linewidth,
        x_label=x_label,
        y_label=y_label,
        figsize=figsize,
        dpi=dpi,
    ).plot(**kwargs)


def plot_loss(
    model: CEBRA,
    label: Optional[Union[str, int, float]] = None,
    ax: Optional[matplotlib.axes.Axes] = None,
    color: str = "magenta",
    linewidth: int = 1,
    x_label: bool = True,
    y_label: bool = True,
    figsize: tuple = (7, 4),
    dpi: float = 100,
    **kwargs,
) -> matplotlib.axes.Axes:
    """Plot the evolution of loss during model training.

    The function makes use of :py:func:`matplotlib.pyplot.plot` and parameters from that function can be provided
    as part of ``kwargs``.

    Args:
        model: The (trained) CEBRA model.
        label: The legend for the loss trace.
        ax: Optional axis to create the plot on.
        color: Line color.
        linewidth: Line width.
        x_label: A boolean that specifies if the x-axis label should be displayed.
        y_label: A boolean that specifies if the y-axis label should be displayed.
        figsize: Figure width and height in inches.
        dpi: Figure resolution.
        kwargs: Optional arguments to customize the plots. See :py:func:`matplotlib.pyplot.plot` documentation for more
            details on which arguments to use.

    Example:

        >>> import cebra
        >>> import numpy as np
        >>> X = np.random.uniform(size=(100, 50))
        >>> y = np.random.uniform(size=(100, 5))
        >>> cebra_model = cebra.CEBRA(max_iterations=10)
        >>> cebra_model.fit(X, y)
        CEBRA(max_iterations=10)
        >>> ax = cebra.plot_loss(cebra_model)

    Returns:
        The axis :py:meth:`matplotlib.axes.Axes.axis` of the plot.

    """

    return _LossPlot(
        model=model,
        label=label,
        axis=ax,
        color=color,
        linewidth=linewidth,
        x_label=x_label,
        y_label=y_label,
        figsize=figsize,
        dpi=dpi,
    ).plot(**kwargs)


def plot_embedding(
    embedding: Union[npt.NDArray, torch.Tensor],
    embedding_labels: Optional[Union[npt.NDArray, torch.Tensor, str]] = "grey",
    ax: Optional[matplotlib.axes.Axes] = None,
    idx_order: Optional[Tuple[int]] = None,
    markersize: float = 0.05,
    alpha: float = 0.4,
    cmap: str = "cool",
    title: str = "Embedding",
    figsize: Tuple[int] = (5, 5),
    dpi: float = 100,
    **kwargs,
) -> matplotlib.axes.Axes:
    """Plot embedding in a 3D or 2D dimensional space.

    If the embedding dimension is equal or higher to 3:

        - If ``idx_order`` is not provided, the plot will be 3D by default.
        - If ``idx_order`` is provided, and it has 3 dimensions, the plot will be 3D;
          if only 2 dimensions are provided, the plot will be 2D.

    If the embedding dimension is equal to 2:

        - If ``idx_order`` is not provided, the plot will be 2D by default.
        - If ``idx_order`` is provided, and it has 3 dimensions, the plot will be 3D;
          if 2 dimensions are provided, the plot will be 2D.

    This assumes that the dimensions provided to ``idx_order`` are within the range of the
    number of dimensions of the embedding (i.e., between 0 and
    :py:attr:`cebra.CEBRA.output_dimension` -1).

    The function makes use of :py:func:`matplotlib.pyplot.scatter`, and parameters from
    that function can be provided as part of ``kwargs``.

    Args:
        embedding: A matrix containing the feature representation computed with CEBRA.
        embedding_labels: The labels used to map the data to color. It can be:

            - A vector that is the same sample size as the embedding, associating a value
              to each sample, either discrete or continuous.
            - A string, either `time`, which will color the embedding based on temporality,
              or a string that can be interpreted as an RGB(A) color, which will display
              the embedding uniformly with that color.

        ax: Optional axis to create the plot on.
        idx_order: A tuple (x, y, z) or (x, y) that maps a dimension in the data to a dimension
            in the 3D/2D embedding. The simplest form is (0, 1, 2) or (0, 1), but one might
            want to plot either those dimensions differently (e.g., (1, 0, 2)) or other
            dimensions from the feature representation (e.g., (2, 4, 5)).

        markersize: The marker size.
        alpha: The marker blending, between 0 (transparent) and 1 (opaque).
        cmap: The Colormap instance or registered colormap name used to map scalar data to colors.
            It will be ignored if `embedding_labels` is set to a valid RGB(A).
        title: The title on top of the embedding.
        figsize: Figure width and height in inches.
        dpi: Figure resolution.
        kwargs: Optional arguments to customize the plots. See :py:func:`matplotlib.pyplot.scatter`
            documentation for more details on which arguments to use.

    Returns:
        The axis :py:meth:`matplotlib.axes.Axes.axis` of the plot.

    Example:
        >>> import cebra
        >>> import numpy as np
        >>> X = np.random.uniform(0, 1, (100, 50))
        >>> y = np.random.uniform(0, 10, (100, 5))
        >>> cebra_model = cebra.CEBRA(max_iterations=10)
        >>> cebra_model.fit(X, y)
        CEBRA(max_iterations=10)
        >>> embedding = cebra_model.transform(X)
        >>> ax = cebra.plot_embedding(embedding, embedding_labels='time')
    """

    return _EmbeddingPlot(
        embedding=embedding,
        embedding_labels=embedding_labels,
        axis=ax,
        idx_order=idx_order,
        markersize=markersize,
        alpha=alpha,
        cmap=cmap,
        title=title,
        figsize=figsize,
        dpi=dpi,
    ).plot(**kwargs)


def plot_consistency(
    scores: Union[npt.NDArray, torch.Tensor, list],
    pairs: Optional[Union[npt.NDArray, list]] = None,
    datasets: Optional[Union[npt.NDArray, list]] = None,
    ax: Optional[matplotlib.axes.Axes] = None,
    cmap: str = "binary",
    text_color: str = "white",
    colorbar_label: Optional[str] = "Consistency score (%)",
    title: Optional[str] = None,
    figsize: Tuple[int] = (5, 5),
    dpi: float = 100,
    **kwargs,
) -> matplotlib.axes.Axes:
    """Plot the consistency matrix from the consistency scores obtained with :py:func:`~.consistency_score`.

    To display the consistency map, ``datasets`` and ``pairs`` must be provided.
    The function is implemented so that it will first try to display the labels as if the
    scores were computed following a between-sujects comparison.
    If only one dataset is present or if the number of datasets doesn't fit with
    the number of scores, then the function will display the labels as if the scores were computed following a
    between-runs comparison. If the number of pairs still doesn't fit with that comparison, then an error is
    raised, asking the user to make sure that the formats of the scores, datasets and pairs are valid.

    .. tip::
        The safer way to use that function is to directly use as inputs the outputs from :py:func:`~.consistency_score`.

    The function makes use of :py:func:`matplotlib.pyplot.imshow` and parameters from that function can be provided as part of ``kwargs``.
    For instance, we recommend that you bound the color bar using ``vmin`` and ``vmax``. The scores are percentages between 0 and 100.

    Args:
        scores: List of consistency scores obtained by comparing a set of CEBRA embeddings using :py:func:`~.consistency_score`.
        pairs: List of the pairs of datasets/runs whose embeddings were compared in ``scores``.
        datasets: List of the datasets whose embeddings were compared in ``scores``. Each dataset is present once only in the list.
        ax: Optional axis to create the plot on.
        cmap: Color map to use to color the m
        text_color: If None, then the ``scores`` values are not displayed, else, it corresponds to the color
            of the values displayed inside the grid.
        colorbar_label: If None, then the color bar is not shown, else, it defines the corresponding title.
        title: The title on top of the confusion matrix.
        figsize: Figure width and height in inches.
        dpi: Figure resolution.
        kwargs: Optional arguments to customize the plots. See :py:func:`matplotlib.pyplot.imshow` documentation for more
            details on which arguments to use.

    Returns:
        The axis :py:meth:`matplotlib.axes.Axes.axis` of the plot.

    Example:

        >>> import cebra
        >>> import numpy as np
        >>> embedding1 = np.random.uniform(0, 1, (1000, 5))
        >>> embedding2 = np.random.uniform(0, 1, (1000, 8))
        >>> labels1 = np.random.uniform(0, 1, (1000, ))
        >>> labels2 = np.random.uniform(0, 1, (1000, ))
        >>> dataset_ids = ["achilles", "buddy"]
        >>> # between-datasets consistency, by aligning on the labels
        >>> scores, pairs, datasets = cebra.sklearn.metrics.consistency_score(
        ...     embeddings=[embedding1, embedding2],
        ...     labels=[labels1, labels2],
        ...     dataset_ids=dataset_ids,
        ...     between="datasets"
        ... )
        >>> ax = cebra.plot_consistency(scores, pairs, datasets, vmin=0, vmax=100)

    """

    return _ConsistencyPlot(
        scores=scores,
        datasets=datasets,
        pairs=pairs,
        axis=ax,
        cmap=cmap,
        text_color=text_color,
        colorbar_label=colorbar_label,
        title=title,
        figsize=figsize,
        dpi=dpi,
    ).plot(**kwargs)


@requires_package_version(matplotlib, "3.6")
def compare_models(
    models: List[CEBRA],
    labels: Optional[List[str]] = None,
    ax: Optional[matplotlib.axes.Axes] = None,
    color: Optional[str] = None,
    cmap: str = "cebra",
    linewidth: int = 1,
    x_label: bool = True,
    y_label: bool = True,
    figsize: tuple = (7, 4),
    dpi: float = 100,
    **kwargs,
) -> matplotlib.axes.Axes:
    """Plot the evolution of loss during model training.

    The function makes use of :py:func:`matplotlib.pyplot.plot` and parameters from that function can be provided
    as part of ``kwargs``.

    Args:
        models: A list of trained CEBRA models.
        labels: A list of labels, associated to each model to be used in the legend of the plot.
        ax: Optional axis to create the plot on.
        color: Line color. If not ``None``, then all the traces for all models will be of the same color.
        cmap: Color map from which to sample uniformly the quantitative list of colors for each trace.
            If ``color`` is different from then that parameter is ignored.
        linewidth: Line width.
        x_label: A boolean that specifies if the x-axis label should be displayed.
        y_label: A boolean that specifies if the y-axis label should be displayed.
        figsize: Figure width and height in inches.
        dpi: Figure resolution.
        kwargs: Optional arguments to customize the plots. See :py:func:`matplotlib.pyplot.plot` documentation for more
            details on which arguments to use.

    Example:

        >>> import cebra
        >>> import numpy as np
        >>> X = np.random.uniform(0, 1, (100, 50))
        >>> y = np.random.uniform(0, 10, (100, 5))
        >>> output_dimensions = [3, 8, 12, 32]
        >>> models, labels = [], []
        >>> for output_dimension in output_dimensions:
        ...     cebra_model = cebra.CEBRA(max_iterations=10,
        ...                               output_dimension=output_dimension)
        ...     cebra_model = cebra_model.fit(X, y)
        ...     models.append(cebra_model)
        ...     labels.append(f"Output dimension: {output_dimension}")
        >>> ax = cebra.compare_models(models, labels)

    Returns:
        The axis of the generated plot. If no ``ax`` argument was specified, it will be created
        by the function and returned here.
    """

    if not isinstance(models, list):
        raise ValueError(f"Invalid list of models, got {type(models)}.")

    for model in models:
        if not isinstance(model, CEBRA):
            raise ValueError(
                f"Invalid list of models, it should only contain CEBRA models, got {model}, of type {type(model)}"
            )
    n_models = len(models)

    # check the color of the traces
    if color is None:
        cebra_map = plt.get_cmap(cmap)
        colors = matplotlib.colors.ListedColormap(
            cebra_map.resampled(n_models)(np.arange(n_models))).colors
    else:
        colors = [color for i in range(n_models)]

    # check the labels
    if labels is None:
        labels = [None for i in range(n_models)]
    else:
        if not isinstance(labels, list):
            raise ValueError(f"Invalid list of labels, got {type(labels)}.")
        if len(labels) != len(models):
            raise ValueError(
                f"Invalid list of labels, it should be the same length as the list of models,"
                f"got {len(labels)}, expected {len(models)}.")
        for label in labels:
            if not (isinstance(label, str) or isinstance(label, int) or
                    isinstance(label, float)):
                raise ValueError(
                    f"Invalid list of labels, it should only contain strs/ints/floats, got {label}, of type {type(label)}"
                )

    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)

    for color, model, label in zip(colors, models, labels):
        ax = plot_loss(
            model,
            color=color,
            label=label,
            linewidth=linewidth,
            x_label=x_label,
            y_label=y_label,
            dpi=dpi,
            ax=ax,
            **kwargs,
        )

    return ax


_register_colormap()



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Some decoders following the ``scikit-learn`` API."""

import abc
from typing import Generator, Optional, Tuple, Union

import numpy as np
import numpy.typing as npt
import sklearn
import sklearn.base
import sklearn.neighbors
import torch

import cebra.helper


class Decoder(abc.ABC, sklearn.base.BaseEstimator):
    """Abstract base class for implementing a decoder."""

    @abc.abstractmethod
    def fit(
        self,
        X: Union[npt.NDArray, torch.Tensor],
        y: Union[npt.NDArray, torch.Tensor],
    ) -> "Decoder":
        """Fit the decoder.

        Args:
            X: The data matrix to decode from.
            y: A 1D array containing the targets.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def predict(self, X: Union[npt.NDArray, torch.Tensor]) -> npt.NDArray:
        """Predict the data ``X``.

        Args:
            X: The data matrix to predict.
        """
        raise NotImplementedError()

    def score(
            self, X: Union[npt.NDArray, torch.Tensor],
            y: Union[npt.NDArray, torch.Tensor]) -> Tuple[float, float, float]:
        """Returns performances of the decoder instance on the provided data ``X``.

        Args:
            X: A data matrix.
            y: The true targets.

        Returns:
            The R2 score on ``X``.
        """
        if isinstance(y, torch.Tensor):
            y = y.numpy()

        prediction = self.predict(X)
        test_score = sklearn.metrics.r2_score(y, prediction)
        return test_score


class KNNDecoder(Decoder):
    """Decoder implementing the k-nearest neighbors vote.

    Note:
        See `sklearn.neighbors.KNeighborsClassifier <https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.KNeighborsClassifier.html>`_
        and `sklearn.neighbors.KNeighborsRegressor <https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.KNeighborsRegressor.html>`_.

    Attributes:
        n_neighbors (int): An integer indicating the K number of neighbors to consider.
        metric (str): The metric to evaluate the KNN decoder's performances.

    Examples:

        >>> from cebra import KNNDecoder
        >>> import numpy as np
        >>> X = np.random.uniform(0, 1, (100, 50))
        >>> y = np.random.uniform(0, 10, (100, 2))
        >>> decoder = KNNDecoder()
        >>> decoder.fit(X, y)
        KNNDecoder()
        >>> score = decoder.score(X, y)

    """

    def __init__(self, n_neighbors: int = 3, metric: Optional[str] = "cosine"):
        self.n_neighbors = n_neighbors
        self.metric = metric

    def fit(
        self,
        X: Union[npt.NDArray, torch.Tensor],
        y: Union[npt.NDArray, torch.Tensor],
    ) -> "KNNDecoder":
        """Fit the KNN decoder.

        Args:
            X: The data matrix.
            y: A 1D array containing the targets.

        Returns:
            ``self``, to allow chaining of operations.
        """

        # Check validity of the target vector
        if len(y) != len(X):
            raise ValueError(
                f"Invalid shape: y and X must have the same number of samples, got y:{len(y)} and X:{len(X)}."
            )

        # Use regression or classification, based on if the targets are continuous or discrete
        if cebra.helper._is_floating(y):
            self.knn = sklearn.neighbors.KNeighborsRegressor(
                n_neighbors=self.n_neighbors, metric=self.metric)
        elif cebra.helper._is_integer(y):
            self.knn = sklearn.neighbors.KNeighborsClassifier(
                n_neighbors=self.n_neighbors, metric=self.metric)
        else:
            raise NotImplementedError(
                f"Invalid type: targets must be either floats or integers, got y:{y.dtype}."
            )

        self.knn.fit(X, y)
        return self

    def predict(
        self, X: Union[npt.NDArray,
                       torch.Tensor]) -> Union[npt.NDArray, torch.Tensor]:
        """Predict the targets for data ``X``.

        Args:
            X: The data matrix.

        Returns:
            A matrix with the prediction for each data sample.
        """
        return self.knn.predict(X)

    def iter_hyperparams() -> Generator[dict, None, None]:
        """Create sets of parameters.

        Note:
            It can be used for parametrized testing.

        Yields:
            A dictionary containing sets of parameters to be used for
            testing.
        """
        for n in np.power(np.arange(1, 6, dtype=int), 2):
            yield dict(n_neighbors=n, metric="cosine")


class L1LinearRegressor(Decoder):
    """A linear model trained with L1 prior as regularizer (aka the Lasso).

    Attributes:
        alpha (float): regularization strength coefficient.

    Note:
        See `sklearn.linear_model.Lasso <https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Lasso.html>`_.

    Examples:

        >>> from cebra import L1LinearRegressor
        >>> import numpy as np
        >>> X = np.random.uniform(0, 1, (100, 50))
        >>> y = np.random.uniform(0, 10, (100, 2))
        >>> decoder = L1LinearRegressor()
        >>> decoder.fit(X, y)
        L1LinearRegressor()
        >>> score = decoder.score(X, y)
    """

    def __init__(self, alpha: float = 1.0):
        self.alpha = alpha

    def fit(
        self,
        X: Union[npt.NDArray, torch.Tensor],
        y: Union[npt.NDArray, torch.Tensor],
    ) -> "L1LinearRegressor":
        """Fit the Lasso regressor.

        Args:
            X: The data matrix.
            y: A 1D array containing the targets.

        Returns:
            ``self``, to allow chaining of operations.
        """
        # Check the targets validity
        if len(y) != len(X):
            raise ValueError(
                f"Invalid shape: y and X must have the same number of samples, got y:{len(y)} and X:{len(X)}."
            )

        if not (cebra.helper._is_integer(y) or cebra.helper._is_floating(y)):
            raise NotImplementedError(
                f"Invalid type: targets must be numeric, got y:{y.dtype}")

        self.model = sklearn.linear_model.Lasso(alpha=self.alpha)
        self.model.fit(X, y)
        return self

    def predict(
        self, X: Union[npt.NDArray,
                       torch.Tensor]) -> Union[npt.NDArray, torch.Tensor]:
        """Predict the targets for data ``X``.

        Args:
            X: The data matrix.

        Returns:
            A matrix with the prediction for each data sample.
        """
        return self.model.predict(X)

    def iter_hyperparams() -> Generator[dict, None, None]:
        """Create sets of parameters.

        Note:
            It can be used for testing.

        Yields:
            A dictionary containing sets of parameters to be used for
            testing.
        """
        for alpha in [0.001, 0.01, 0.1, 1, 10, 100]:
            yield dict(alpha=alpha)



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Ephys neural and behavior data used for the monkey reaching experiment.

References:
    * Chowdhury, Raeed H., Joshua I. Glaser, and Lee E. Miller.
      "Area 2 of primary somatosensory cortex encodes kinematics of the whole arm."
      Elife 9 (2020).
    * Chowdhury, Raeed; Miller, Lee (2022)
      Area2 Bump: macaque somatosensory area 2 spiking activity during reaching
      with perturbations (Version 0.220113.0359) [Data set].
      `DANDI archive <https://doi.org/10.48324/dandi.000127/0.220113.0359>`_
    * Pei, Felix, et al.
      "Neural Latents Benchmark'21: Evaluating latent variable models of neural
      population activity." arXiv preprint arXiv:2109.04463 (2021).
"""

import pathlib
from typing import Union

import joblib as jl
import numpy as np
import torch

import cebra.data
from cebra.datasets import get_datapath
from cebra.datasets import register

_DEFAULT_DATADIR = get_datapath()


def _load_data(
    path: Union[str, pathlib.Path] = None,
    session: str = "active",
    split: str = "train",
):
    """Load and preprocess neural and behavior data of monkey reaching task from NWBDataset.

    Ephys and behavior recording from -100ms and 500ms from the movement onset in 1ms bin size.
    Neural recording is smoothened with Gaussian kernel with 40ms std.
    The behavior labels include trial types, target directions and the x,y hand positions.

    Args:
        path: The path to the nwb file.
        session: The session type to load among 'active', 'passive' and 'all'.
        split: The split to load among 'train', 'valid', 'test' and 'all'.

    """

    if path is None:
        path = pathlib.Path(
            _DEFAULT_DATADIR
        ) / "s1_reaching" / "sub-Han_desc-train_behavior+ecephys.nwb"
    else:
        path = pathlib.Path(path)

    try:
        from nlb_tools.nwb_interface import NWBDataset
    except ImportError:
        raise ImportError(
            "Could not import the nlb_tools package required for data loading "
            "the raw reaching datasets in NWB format. "
            "If required, you can install the dataset by running "
            "pip install nlb_tools."
            # NOTE(stes): Install nlb_tools manually for now, also see
            # note in setup.cfg
            # or installing cebra with the [datasets] "
            #"dependencies: pip install 'cebra[datasets]'")
        )

    def _get_info(trial_info, data):
        passive = []
        direction = []
        direction_actpas = []
        for index, trial in trial_info.iterrows():
            if trial["ctr_hold_bump"]:
                passive.append(True)
                direction.append(int(trial["bump_dir"] / 45))
                direction_actpas.append(int(trial["bump_dir"] / 45) + 8)
            else:
                passive.append(False)
                direction.append(int(trial["cond_dir"] / 45))
                direction_actpas.append(int(trial["cond_dir"] / 45))
            spikes = data["spikes_smth_40"].to_numpy()
            velocity = data["hand_vel"].to_numpy()
            position = data["hand_pos"].to_numpy()

        return {
            "spikes": spikes,
            "vel": velocity,
            "pos": position,
            "passive": np.array(passive),
            "movement_dir": np.array(direction),
            "movement_dir_actpas": np.array(direction_actpas),
            "num_trials": len(trial_info),
            "trial_len": int(len(spikes) / len(trial_info)),
        }

    dataset = NWBDataset(path, split_heldout=False)
    dataset.smooth_spk(40, name="smth_40")

    if split == "train":
        split_mask = dataset.trial_info.split == "train"

    elif split == "all":
        split_mask = dataset.trial_info.split != "none"

    elif split == "valid" or split == "test":
        split_mask = dataset.trial_info.split == "val"
    else:
        raise ValueError("--split argument should be train, valid, test or all")

    if session == "active":
        session_mask = ~dataset.trial_info.ctr_hold_bump
    elif session == "passive":
        session_mask = dataset.trial_info.ctr_hold_bump
    elif session == "all":
        session_mask = True
    else:
        raise ValueError("--session argument should be active, passive or all")

    mask = split_mask & session_mask

    trials = dataset.make_trial_data(align_field="move_onset_time",
                                     align_range=(-100, 500),
                                     ignored_trials=~mask)
    trial_info = dataset.trial_info[mask]

    if split == "valid":
        data_dic = _get_info(trial_info[:mask.sum() // 2],
                             trials[:len(trials) // 2])
    elif split == "test":
        data_dic = _get_info(trial_info[mask.sum() // 2:],
                             trials[len(trials) // 2:])
    else:
        data_dic = _get_info(trial_info, trials)

    return data_dic


monkey_reaching_urls = {
    "all_all.jl": {
        "url":
            "https://figshare.com/ndownloader/files/41668764?private_link=6fa4ee74a8f465ec7914",
        "checksum":
            "dea556301fa4fafa86e28cf8621cab5a"
    },
    "all_train.jl": {
        "url":
            "https://figshare.com/ndownloader/files/41668752?private_link=6fa4ee74a8f465ec7914",
        "checksum":
            "e280e4cd86969e6fd8bfd3a8f402b2fe"
    },
    "all_test.jl": {
        "url":
            "https://figshare.com/ndownloader/files/41668761?private_link=6fa4ee74a8f465ec7914",
        "checksum":
            "25d3ff2c15014db8b8bf2543482ae881"
    },
    "all_valid.jl": {
        "url":
            "https://figshare.com/ndownloader/files/41668755?private_link=6fa4ee74a8f465ec7914",
        "checksum":
            "8cd25169d31f83ae01b03f7b1b939723"
    },
    "active_all.jl": {
        "url":
            "https://figshare.com/ndownloader/files/41668776?private_link=6fa4ee74a8f465ec7914",
        "checksum":
            "c626acea5062122f5a68ef18d3e45e51"
    },
    "active_train.jl": {
        "url":
            "https://figshare.com/ndownloader/files/41668770?private_link=6fa4ee74a8f465ec7914",
        "checksum":
            "72a48056691078eee22c36c1992b1d37"
    },
    "active_test.jl": {
        "url":
            "https://figshare.com/ndownloader/files/41668773?private_link=6fa4ee74a8f465ec7914",
        "checksum":
            "35b7e060008a8722c536584c4748f2ea"
    },
    "active_valid.jl": {
        "url":
            "https://figshare.com/ndownloader/files/41668767?private_link=6fa4ee74a8f465ec7914",
        "checksum":
            "dd58eb1e589361b4132f34b22af56b79"
    },
    "passive_all.jl": {
        "url":
            "https://figshare.com/ndownloader/files/41668758?private_link=6fa4ee74a8f465ec7914",
        "checksum":
            "bbb1bc9d8eec583a46f6673470fc98ad"
    },
    "passive_train.jl": {
        "url":
            "https://figshare.com/ndownloader/files/41668743?private_link=6fa4ee74a8f465ec7914",
        "checksum":
            "f22e05a69f70e18ba823a0a89162a45c"
    },
    "passive_test.jl": {
        "url":
            "https://figshare.com/ndownloader/files/41668746?private_link=6fa4ee74a8f465ec7914",
        "checksum":
            "42453ae3e4fd27d82d297f78c13cd6b7"
    },
    "passive_valid.jl": {
        "url":
            "https://figshare.com/ndownloader/files/41668749?private_link=6fa4ee74a8f465ec7914",
        "checksum":
            "2dcc10c27631b95a075eaa2d2297bb4a"
    }
}


@register("area2-bump")
class Area2BumpDataset(cebra.data.SingleSessionDataset):
    """Base dataclass to generate monkey reaching datasets.

    Ephys and behavior recording from -100ms and 500ms from the movement
    onset in 1ms bin size.
    Neural recording is smoothened with Gaussian kernel with 40ms std.
    The behavior labels can include trial types, target directions and the
    x,y hand positions.
    After initialization of the dataset, split method can splits the data
    into 'train', 'valid' and 'test' split.

    Args:
        path: The path to the directory where the preloaded data is.
        session: The trial type. Choose between 'active', 'passive',
            'all', 'active-passive'.

    """

    def __init__(self,
                 path: str = get_datapath("monkey_reaching_preload_smth_40/"),
                 session: str = "active",
                 download=True):
        super().__init__()
        self.path = path
        self.download = download
        self.session = session
        if session == "active-passive":
            self.load_session = "all"
        else:
            self.load_session = session

        super().__init__(
            download=self.download,
            data_url=monkey_reaching_urls[f"{self.load_session}_all.jl"]["url"],
            data_checksum=monkey_reaching_urls[f"{self.load_session}_all.jl"]
            ["checksum"],
            location=self.path,
            file_name=f"{self.load_session}_all.jl",
        )

        self.data = jl.load(
            pathlib.Path(self.path) / f"{self.load_session}_all.jl")
        self._post_load()

    def split(self, split):
        """Split the dataset.

        The train trials are the same as one defined in Neural Latent
        Benchmark (NLB) Dataset.
        The half of the valid trials defined in NLBDataset is used as
        the valid set and the other half is used as the test set.

        Args:
            split: The split. It can be either `all`, `train`, `valid`, `test`.

        """

        super().__init__(
            download=self.download,
            data_url=monkey_reaching_urls[f"{self.load_session}_{split}.jl"]
            ["url"],
            data_checksum=monkey_reaching_urls[
                f"{self.load_session}_{split}.jl"]["checksum"],
            location=self.path,
            file_name=f"{self.load_session}_{split}.jl",
        )
        self.data = jl.load(
            pathlib.Path(self.path) / f"{self.load_session}_{split}.jl")
        self._post_load()

    def _post_load(self):
        """Read and assign neural and behavior recording into the class attributes."""

        self.trial_len = int(self.data["trial_len"])
        self.num_trials = int(self.data["num_trials"])
        self.neural = torch.from_numpy(self.data["spikes"]).float()
        self.trial_ids = np.concatenate(
            [[n] * self.trial_len for n in range(self.num_trials)])
        self.trial_borders = [
            self.trial_len * i for i in range(self.num_trials + 1)
        ]
        self.trial_indices = np.concatenate(
            [np.arange(self.trial_len) for n in range(self.num_trials)])

        self.passive = torch.from_numpy(
            np.concatenate([[t] * self.trial_len for t in self.data["passive"]
                           ])).long()
        self.pos = torch.from_numpy(self.data["pos"]).float()
        self.vel = torch.from_numpy(self.data["vel"]).float()
        self.target = torch.from_numpy(
            np.concatenate([
                [t] * self.trial_len for t in self.data["movement_dir"]
            ])).long()
        self.target_actpas = torch.from_numpy(
            np.concatenate([
                [t] * self.trial_len for t in self.data["movement_dir_actpas"]
            ])).long()
        self.trial_indices = torch.from_numpy(self.trial_indices).float()

    @property
    def input_dimension(self):
        return self.neural.size(1)

    def __len__(self):
        return len(self.neural)

    @property
    def discrete_index(self):
        return self.passive

    @property
    def continuous_index(self):
        return self.pos

    def __repr__(self):
        return f"MonkeyArea2BumpDataset(name: Discrete active/passive & " \
               f"continuous hand position, shape: {self.neural.shape})"

    def __getitem__(self, index):
        index = self.expand_index_in_trial(index,
                                           trial_ids=self.trial_ids,
                                           trial_borders=self.trial_borders)
        return self.neural[index].transpose(2, 1)


@register("area2-bump-shuffled")
class Area2BumpShuffledDataset(Area2BumpDataset):
    """Base dataclass to generate shuffled monkey reaching datasets.

    Ephys and behavior recording from -100ms and 500ms from the movement
    onset in 1ms bin size.
    Neural recording is smoothened with Gaussian kernel with 40ms std.
    The shuffled behavior labels can include trial types, target directions
    and the x,y hand positions.

    After initialization of the dataset, split method can splits the data
    into 'train', 'valid' and 'test' split.

    Args:
        path: The path to the directory where the preloaded data is.
        session: The trial type. Choose between 'active', 'passive', 'all',
            'active-passive'.

    """

    def _post_load(self):
        rng = np.random.Generator(np.random.PCG64(1))

        self.trial_len = int(self.data["trial_len"])
        self.num_trials = int(self.data["num_trials"])
        self.neural = torch.from_numpy(self.data["spikes"]).float()
        self.trial_ids = np.concatenate(
            [[n] * self.trial_len for n in range(self.num_trials)])
        self.trial_borders = [
            self.trial_len * i for i in range(self.num_trials + 1)
        ]
        self.trial_indices = np.concatenate(
            [np.arange(self.trial_len) for n in range(self.num_trials)])

        shuffle_index = np.arange(len(self.neural))
        rng.shuffle(shuffle_index)

        self.passive = torch.from_numpy(
            np.concatenate([[t] * self.trial_len for t in self.data["passive"]
                           ])).long()
        self.passive_shuffled = self.passive[shuffle_index]
        self.pos = torch.from_numpy(self.data["pos"]).float()
        self.pos_shuffled = self.pos[shuffle_index]
        self.target = torch.from_numpy(
            np.concatenate([
                [t] * self.trial_len for t in self.data["movement_dir"]
            ])).long()
        self.target_actpas = torch.from_numpy(
            np.concatenate([
                [t] * self.trial_len for t in self.data["movement_dir_actpas"]
            ])).long()
        self.target_shuffled = self.target[shuffle_index]
        self.trial_indices = torch.from_numpy(self.trial_indices).float()


def _create_area2_dataset():
    """Register the monkey reaching datasets of different trial types, behavior labels.

    The trial types are 'active', 'passive', 'all' and 'active-passive'.
    The 'active-passive' type distinguishes movement direction between active, passive
    (0-7 for active and 8-15 for passive) and 'all' does not (0-7).

    """

    PATH = pathlib.Path(_DEFAULT_DATADIR) / "monkey_reaching_preload_smth_40"
    for session_type in ["active", "passive", "active-passive", "all"]:

        @register(f"area2-bump-pos-{session_type}")
        class DatasetV1(Area2BumpDataset):
            """Monkey reaching dataset with hand position labels.

            The dataset loads continuous x,y hand position as behavior labels.
            For the 'active-passive' trial type, it additionally loads discrete binary
            label of active(0)/passive(1).

            Args:
                path: The path to the directory where the preloaded data is.
                session: The trial type. Choose between 'active', 'passive', 'all',
                    'active-passive'.

            """

            def __init__(self, path=PATH, session=session_type):
                super().__init__(path=path, session=session)

            @property
            def discrete_index(self):
                if self.session == "active-passive":
                    return self.passive
                else:
                    return None

            @property
            def continuous_index(self):
                return self.pos

        @register(f"area2-bump-target-{session_type}")
        class DatasetV2(Area2BumpDataset):
            """Monkey reaching dataset with target direction labels.

            The dataset loads discrete target direction (0-7) as behavior labels.

            Args:
                path: The path to the directory where the preloaded data is.
                session: The trial type. Choose between 'active', 'passive', 'all',
                    'active-passive'.

            """

            def __init__(self, path=PATH, session=session_type):
                super().__init__(path=path, session=session)

            @property
            def discrete_index(self):
                if self.session == "active-passive":
                    return self.target_actpas
                else:
                    return self.target

            @property
            def continuous_index(self):
                return None

        @register(f"area2-bump-posdir-{session_type}")
        class DatasetV3(Area2BumpDataset):
            """Monkey reaching dataset with hand position labels and discrete target labels.

            The dataset loads continuous x,y hand position and discrete target labels (0-7)
            as behavior labels.
            For active-passive type, the discrete target labels 0-7 for active and 8-16 for
            passive are loaded.

            Args:
                path: The path to the directory where the preloaded data is.
                session: The trial type. Choose between 'active', 'passive', 'all',
                'active-passive'.

            """

            def __init__(self, path=PATH, session=session_type):
                super().__init__(path=path, session=session)

            @property
            def discrete_index(self):
                return self.target

            @property
            def continuous_index(self):
                return self.pos


_create_area2_dataset()


def _create_area2_shuffled_dataset():
    """Register the shuffled monkey reaching datasets of different trial types, behavior labels.

    The trial types are 'active' and 'active-passive'.
    The behavior labels are randomly shuffled and the trial types are shuffled
    in case of 'shuffled-trial' datasets.

    """

    PATH = pathlib.Path(_DEFAULT_DATADIR) / "monkey_reaching_preload_smth_40"
    for session_type in ["active", "active-passive"]:

        @register(f"area2-bump-pos-{session_type}-shuffled-trial")
        class DatasetV4(Area2BumpShuffledDataset):
            """Monkey reaching dataset with the shuffled trial type.

            The dataset loads the discrete binary trial type label active(0)/passive(1)
            in randomly shuffled order.

            Args:
                path: The path to the directory where the preloaded data is.
                session: The trial type. Choose between 'active', 'passive', 'all',
                    'active-passive'.

            """

            def __init__(self, path=PATH, session=session_type):
                super().__init__(path=path, session=session)

            @property
            def discrete_index(self):
                if self.session == "active-passive":
                    return self.passive_shuffled
                else:
                    return None

            @property
            def continuous_index(self):
                return self.pos

        @register(f"area2-bump-pos-{session_type}-shuffled-position")
        class DatasetV5(Area2BumpShuffledDataset):
            """Monkey reaching dataset with the shuffled hand position.

            The dataset loads continuous x,y hand position in randomly shuffled order.
            For the 'active-passive' trial type, it additionally loads discrete binary label
            of active(0)/passive(1).

            Args:
                path: The path to the directory where the preloaded data is.
                session: The trial type. Choose between 'active', 'passive', 'all',
                    'active-passive'.

            """

            def __init__(self, path=PATH, session=session_type):
                super().__init__(path=path, session=session)

            @property
            def discrete_index(self):
                if self.session == "active-passive":
                    return self.passive
                else:
                    return None

            @property
            def continuous_index(self):
                return self.pos_shuffled

        @register(f"area2-bump-target-{session_type}-shuffled")
        class DatasetV6(Area2BumpShuffledDataset):
            """Monkey reaching dataset with the shuffled hand position.

            The dataset loads discrete target direction (0-7 for active and 0-15 for active-passive)
            in randomly shuffled order.

            Args:
                path: The path to the directory where the preloaded data is.
                session: The trial type. Choose between 'active', 'passive', 'all',
                    'active-passive'.

            """

            def __init__(self, path=PATH, session=session_type):
                super().__init__(path=path, session=session)

            @property
            def discrete_index(self):
                if self.session == "active-passive":
                    return self.passive
                else:
                    return self.target_shuffled

            @property
            def continuous_index(self):
                return None


_create_area2_shuffled_dataset()
